# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""CommandContainer module

Classes:
- CommandContainer, a special mixin class to be used with
Hardware Objects. It defines a container
for command launchers and channels (see Command package).
- C*Object, command launcher & channel base class
"""

from __future__ import absolute_import

import logging
import weakref
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

from mxcubecore.dispatcher import dispatcher

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"

PROCEDURE_COMMAND_T = "CONTROLLER"
TWO_STATE_COMMAND_T = "INOUT"

ARGUMENT_TYPE_LIST = "List"
ARGUMENT_TYPE_JSON_SCHEMA = "JSONSchema"


class ConnectionError(Exception):
    """General Connection Error"""


class CommandObject:
    def __init__(self, name: str, username: Optional[str] = None, **kwargs) -> None:
        """
        Args:
            name (str): Name.
            username (Optional[str], optional): User name. Defaults to None.
        """
        self._name: str = name
        self._username: Union[str, None] = username
        self._arguments: Union[List[Tuple[str, str, Any, Any]], Any] = []
        self._combo_arguments_items: Dict[str, Any] = {}

        self.type: str = PROCEDURE_COMMAND_T
        self.argument_type: str = ARGUMENT_TYPE_LIST

    def set_argument_json_schema(self, json_schema_str: Any) -> None:
        """Set the JSON Schema.

        Args:
            json_schema_str (Any): JSON Schema.
        """
        self.argument_type = ARGUMENT_TYPE_JSON_SCHEMA
        self._arguments = json_schema_str

    def name(self) -> str:
        """Get command name.

        Returns:
            str: Command name.
        """
        return self._name

    def connect(self, signal_name: str, callable_func: Callable) -> None:
        """Connect to signal.

        Args:
            signal_name (str): Signal name.
            callable_func (Callable): Connection method.
        """
        try:
            dispatcher.disconnect(callable_func, signal_name, self)
        except Exception:
            pass
        dispatcher.connect(callable_func, signal_name, self)

    def emit(self, signal: str, *args) -> None:
        """Emit signal message.

        Args:
            signal (str): Signal name.
            *args (tuple): Message arguments.
        """
        signal = str(signal)

        if len(args) == 1:
            if isinstance(args[0], tuple):
                args = args[0]

        dispatcher.send(signal, self, *args)

    def add_argument(
        self,
        argName: str,
        argType: str,
        combo_items: Optional[Any] = None,
        onchange: Optional[Any] = None,
        valuefrom: Optional[Any] = None,
    ) -> None:
        """Add command argument.

        Args:
            argName (str): Name.
            argType (str): Type.
            combo_items (Optional[Any], optional): Combo items. Defaults to None.
            onchange (Optional[Any], optional): On change. Defaults to None.
            valuefrom (Optional[Any], optional): Value from. Defaults to None.
        """
        arg_names = [arg[0] for arg in self._arguments]
        if argName not in arg_names:
            self._arguments.append((argName, argType.lower(), onchange, valuefrom))
        if combo_items is not None:
            self._combo_arguments_items[argName] = combo_items

    def get_arguments(self) -> Union[List[Tuple[str, str, Any, Any]], Any]:
        """Get command arguments.

        Returns:
            Union[List[Tuple[str, str, Any, Any]], Any]: Command arguments.
        """
        return self._arguments

    def get_combo_argument_items(self, argName: str) -> Any:
        """Get combo argument items.

        Args:
            argName (str): Combo argument name.

        Returns:
            Any: Combo argument value.
        """
        return self._combo_arguments_items[argName]

    def userName(self) -> str:
        """Get user name.

        Returns:
            str: User name.
        """
        return self._username or str(self.name())

    def is_connected(self) -> bool:
        """Check if signal is connected.

        Returns:
            bool: True if connected, else False.
        """
        return False


class ChannelObject:
    def __init__(self, name: str, username: Optional[str] = None, **kwargs) -> None:
        """
        Args:
            name (str): Name.
            username (Optional[str], optional): User name. Defaults to None.
        """
        self._name: str = name
        self._username: Union[str, None] = username
        self._attributes: Dict[str, Any] = kwargs
        self._on_change: Union[
            Tuple[str, weakref.ref],
            None,
        ] = None
        self.__first_update: bool = True

    def name(self) -> str:
        """Get channel name.

        Returns:
            str: Channel name.
        """
        return self._name

    def connect_signal(self, signalName: str, callableFunc: Callable) -> None:
        """Connect signal.

        Args:
            signalName (str): Signal name.
            callableFunc (Callable): Connection method.
        """
        try:
            dispatcher.disconnect(callableFunc, signalName, self)
        except Exception:
            pass
        dispatcher.connect(callableFunc, signalName, self)

    def disconnect_signal(self, signalName: str, callableFunc: Callable) -> None:
        """Disconnect signal.

        Args:
            signalName (str): Signal name.
            callableFunc (Callable): Disconnection method.
        """
        try:
            dispatcher.disconnect(callableFunc, signalName, self)
        except Exception:
            pass

    def connect_notify(self, signal: str) -> None:
        """Connection notifier.

        Args:
            signal (str): Signal name.
        """
        if signal == "update" and self.is_connected():
            self.emit(signal, self.get_value())

    def emit(self, signal: str, *args) -> None:
        """Emit signal message.

        Args:
            signal (str): Signal name.
            *args (tuple): Message arguments.
        """
        signal = str(signal)

        if len(args) == 1:
            if isinstance(args[0], tuple):
                args = args[0]

        dispatcher.send(signal, self, *args)

    def userName(self) -> str:
        """Get user name.

        Returns:
            str: User name.
        """
        return self._username or str(self.name())

    def is_connected(self) -> bool:
        """Check if signal is connected.

        Returns:
            bool: True if connected, else False.
        """
        return False

    def update(self, value: Any) -> None:
        """Update command object.

        Args:
            value (Any): Updated value.
        """
        if self.__first_update:
            self.__first_update = False
            return

        if self._on_change is not None:
            cmd, container_ref = self._on_change
            container: "CommandContainer" = container_ref()
            if container is not None:
                cmdobj = container.get_command_object(cmd)
                if cmdobj is not None:
                    cmdobj(value)

    def get_value(self, force: bool = False):
        """Get channel value.

        Args:
            force (bool, optional): Force get value. Defaults to False.

        Raises:
            NotImplementedError: If method has not been implemented for this object.
        """
        # NBNB INCONSISTENT. funcxtion signature matches only
        # Tine and Mockup, but is inconsistent with other subclasses
        raise NotImplementedError


class CommandContainer:
    """Mixin class for generic command and channel containers"""

    def __init__(self) -> None:
        self.__commands: Dict[str, CommandObject] = {}
        self.__channels: Dict[str, ChannelObject] = {}
        self.__commands_to_add: List[Tuple[Dict[str, Any], Union[str, None]]] = []
        self.__channels_to_add: List[Tuple[Dict[str, Any], str]] = []

    def __getattr__(self, attr: str) -> CommandObject:
        try:
            return self.__dict__["_CommandContainer__commands"][attr]
        except KeyError:
            raise AttributeError(attr)

    def get_channel_object(
        self,
        channel_name: str,
        optional: bool = False,
    ) -> Union[ChannelObject, None]:
        """Get channel.

        Args:
            channel_name (str): Channel name.
            optional (bool, optional): If a missing channel should be logged
            as an error. Defaults to False.

        Returns:
            Union[ChannelObject, None]: Channel object or None if not found.
        """
        channel = self.__channels.get(channel_name)
        if channel is None and not optional:
            msg = "%s: Unable to get channel %s" % (self.name(), channel_name)
            logging.getLogger("user_level_log").error(msg)
            # raise Exception(msg)
        return channel

    def get_channel_names_list(self) -> List[str]:
        """Get a list of all channel names.

        Returns:
            List[str]: Channel names.
        """
        return list(self.__channels.keys())

    def add_channel(
        self,
        attributes_dict: Dict[str, Any],
        channel: str,
        add_now: bool = True,
    ) -> Union[ChannelObject, None]:
        """Add channel.

        Args:
            attributes_dict (Dict[str, Any]): Channel attributes.
            channel (str): Channel name.
            add_now (bool, optional): Whether the channel should be added now.
            Defaults to True.

        Raises:
            ConnectionError: If a connection error occured while adding the channel.

        Returns:
            Union[ChannelObject, None]: Channel object or None if adding later.
        """
        if not add_now:
            self.__channels_to_add.append((attributes_dict, channel))
            return
        channel_name: str = attributes_dict["name"]
        channel_type: str = attributes_dict["type"]
        channel_on_change: Union[Any, None] = attributes_dict.get("onchange", None)
        if channel_on_change is not None:
            del attributes_dict["onchange"]
        channel_value_from: Union[Any, None] = attributes_dict.get("valuefrom", None)
        if channel_value_from is not None:
            del attributes_dict["valuefrom"]
        channel_value_from = attributes_dict.get("valuefrom", None)
        del attributes_dict["name"]
        del attributes_dict["type"]

        new_channel: Union[ChannelObject, None] = None
        if self.__channels.get(channel_name) is not None:
            return self.__channels[channel_name]

        if channel_type.lower() == "spec":
            if "version" not in attributes_dict:
                try:
                    attributes_dict["version"] = self.specversion
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Spec import SpecChannel

                new_channel = SpecChannel(channel_name, channel, **attributes_dict)
            except Exception:
                logging.getLogger().error(
                    "%s: cannot add channel %s (hint: check attributes)",
                    self.name(),
                    channel_name,
                )
        elif channel_type.lower() == "taco":
            if "taconame" not in attributes_dict:
                try:
                    attributes_dict["taconame"] = self.taconame
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Taco import TacoChannel

                new_channel = TacoChannel(channel_name, channel, **attributes_dict)
            except Exception:
                logging.getLogger().error(
                    "%s: cannot add channel %s (hint: check attributes)",
                    self.name(),
                    channel_name,
                )
        elif channel_type.lower() == "tango":
            if "tangoname" not in attributes_dict:
                try:
                    attributes_dict["tangoname"] = self.tangoname
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Tango import TangoChannel

                new_channel = TangoChannel(channel_name, channel, **attributes_dict)
            except ConnectionError:
                logging.getLogger().error(
                    "%s: could not connect to device server %s (hint: is it running ?)",
                    self.name(),
                    attributes_dict["tangoname"],
                )
                raise ConnectionError
            except Exception:
                logging.getLogger().exception(
                    "%s: cannot add channel %s (hint: check attributes)",
                    self.name(),
                    channel_name,
                )
        elif channel_type.lower() == "exporter":
            if "exporter_address" not in attributes_dict:
                try:
                    attributes_dict["exporter_address"] = self.exporter_address
                except AttributeError:
                    pass
            host, port = attributes_dict["exporter_address"].split(":")

            try:
                attributes_dict["address"] = host
                attributes_dict["port"] = int(port)
                del attributes_dict["exporter_address"]

                from mxcubecore.Command.Exporter import ExporterChannel

                new_channel = ExporterChannel(channel_name, channel, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    "%s: cannot add exporter channel %s (hint: check attributes)",
                    self.name(),
                    channel_name,
                )
        elif channel_type.lower() == "epics":
            try:
                from mxcubecore.Command.Epics import EpicsChannel

                new_channel = EpicsChannel(channel_name, channel, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    "%s: cannot add EPICS channel %s (hint: check PV name)",
                    self.name(),
                    channel_name,
                )
        elif channel_type.lower() == "tine":
            if "tinename" not in attributes_dict:
                try:
                    attributes_dict["tinename"] = self.tine_name
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Tine import TineChannel

                new_channel = TineChannel(channel_name, channel, **attributes_dict)
            except Exception:
                logging.getLogger("HWR").exception(
                    "%s: cannot add TINE channel %s (hint: check attributes)",
                    self.name(),
                    channel_name,
                )

        elif channel_type.lower() == "sardana":
            if "taurusname" not in attributes_dict:
                try:
                    attributes_dict["taurusname"] = self.taurusname
                except AttributeError:
                    pass
            uribase = attributes_dict["taurusname"]

            try:
                from mxcubecore.Command.Sardana import SardanaChannel

                logging.getLogger().debug(
                    "Creating a sardanachannel - %s / %s / %s",
                    self.name(),
                    channel_name,
                    str(attributes_dict),
                )
                new_channel = SardanaChannel(
                    channel_name, channel, uribase=uribase, **attributes_dict
                )
                logging.getLogger().debug("Created")
            except Exception:
                logging.getLogger().exception(
                    "%s: cannot add SARDANA channel %s (hint: check PV name)",
                    self.name(),
                    channel_name,
                )

        elif channel_type.lower() == "mockup":
            if "default_value" not in attributes_dict:
                try:
                    attributes_dict["default_value"] = float(self.default_value)
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Mockup import MockupChannel

                new_channel = MockupChannel(channel_name, channel, **attributes_dict)
            except Exception:
                logging.getLogger("HWR").exception(
                    "%s: cannot add Mockup channel %s (hint: check attributes)",
                    self.name(),
                    channel_name,
                )

        if new_channel is not None:
            if channel_on_change is not None:
                new_channel._on_change = (channel_on_change, weakref.ref(self))
            else:
                new_channel._on_change = None
            if channel_value_from is not None:
                new_channel._valuefrom = (channel_value_from, weakref.ref(self))
            else:
                new_channel._valuefrom = None

            self.__channels[channel_name] = new_channel

            return new_channel
        else:
            logging.getLogger().exception("Channel is None")

    def set_channel_value(self, channel_name: str, value: Any) -> None:
        """Set channel value.

        Args:
            channel_name (str): Channel name.
            value (Any): Value to set.
        """
        self.__channels[channel_name].set_value(value)

    def get_channel_value(self, channel_name: str) -> Any:
        """Get channel value.

        Args:
            channel_name (str): Channel name.

        Returns:
            Any: Channel value.
        """
        return self.__channels[channel_name].get_value()

    def get_channels(self) -> Generator[ChannelObject, None, None]:
        """Get object channels.

        Yields:
            Generator[ChannelObject, None, None]: Object channels.
        """
        for chan in self.__channels.values():
            yield chan

    def get_command_object(self, cmd_name: str) -> Union[CommandObject, None]:
        """Get command object.

        Args:
            cmd_name (str): Command name.

        Returns:
            Union[CommandObject, None]: Command object or None if not found.
        """
        try:
            return self.__commands.get(cmd_name)
        except Exception as e:
            return None

    def get_commands(self) -> Generator[CommandObject, None, None]:
        """Get object commands.

        Yields:
            Generator[CommandObject, None, None]: Command objects.
        """
        for cmd in self.__commands.values():
            yield cmd

    def get_command_names_list(self) -> List[str]:
        """Get list of command names.

        Returns:
            List[str]: Command names.
        """
        return list(self.__commands.keys())

    def add_command(
        self,
        arg1: Dict[str, Any],
        arg2: Optional[str] = None,
        add_now: bool = True,
    ) -> Union[CommandObject, None]:
        """Add command.

        Args:
            arg1 (Dict[str, Any]): Command attributes.
            arg2 (Optional[str], optional): Command name. Defaults to None.
            add_now (bool, optional): Whether to add command now. Defaults to True.

        Raises:
            ConnectionError: If a connection error occured while adding the command.

        Returns:
            Union[CommandObject, None]: Command object or None if adding later.
        """
        if not add_now:
            self.__commands_to_add.append((arg1, arg2))
            return
        new_command: Union[CommandObject, None] = None

        cmd_name: str
        cmd_type: str
        cmd: Union[str, None]
        if isinstance(arg1, dict):
            attributes_dict = arg1
            cmd = arg2

            cmd_name = attributes_dict["name"]
            cmd_type = attributes_dict["type"]
            del attributes_dict["name"]
            del attributes_dict["type"]
        else:
            attributes_dict = {}
            attributes_dict.update(arg1.get_properties())

            try:
                cmd_name = attributes_dict["name"]
                cmd_type = attributes_dict["type"]
                cmd = attributes_dict["toexecute"]
            except KeyError as err:
                logging.getLogger().error(
                    '%s: cannot add command: missing "%s" property',
                    self.name(),
                    err.args[0],
                )
                return
            else:
                del attributes_dict["name"]
                del attributes_dict["type"]
                del attributes_dict["toexecute"]

        if cmd_type.lower() == "spec":
            if "version" not in attributes_dict:
                try:
                    attributes_dict["version"] = self.specversion
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Spec import SpecCommand

                new_command = SpecCommand(cmd_name, cmd, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    '%s: could not add command "%s" (hint: check command attributes)',
                    self.name(),
                    cmd_name,
                )
        elif cmd_type.lower() == "taco":
            if "taconame" not in attributes_dict:
                try:
                    attributes_dict["taconame"] = self.taconame
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Taco import TacoCommand

                new_command = TacoCommand(cmd_name, cmd, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    '%s: could not add command "%s" (hint: check command attributes)',
                    self.name(),
                    cmd_name,
                )
        elif cmd_type.lower() == "tango":
            if "tangoname" not in attributes_dict:
                try:
                    attributes_dict["tangoname"] = self.tangoname
                except AttributeError:
                    pass
            try:
                from mxcubecore.Command.Tango import TangoCommand

                new_command = TangoCommand(cmd_name, cmd, **attributes_dict)
            except ConnectionError:
                logging.getLogger().error(
                    "%s: could not connect to device server %s (hint: is it running ?)",
                    self.name(),
                    attributes_dict["tangoname"],
                )
                raise ConnectionError
            except Exception:
                logging.getLogger().exception(
                    '%s: could not add command "%s" (hint: check command attributes)',
                    self.name(),
                    cmd_name,
                )

        elif cmd_type.lower() == "exporter":
            if "exporter_address" not in attributes_dict:
                try:
                    attributes_dict["exporter_address"] = self.exporter_address
                except AttributeError:
                    pass
            host, port = attributes_dict["exporter_address"].split(":")

            try:
                attributes_dict["address"] = host
                attributes_dict["port"] = int(port)
                del attributes_dict["exporter_address"]

                from mxcubecore.Command.Exporter import ExporterCommand

                new_command = ExporterCommand(cmd_name, cmd, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    "%s: cannot add command %s (hint: check attributes)",
                    self.name(),
                    cmd_name,
                )
        elif cmd_type.lower() == "epics":
            try:
                from mxcubecore.Command.Epics import EpicsCommand

                new_command = EpicsCommand(cmd_name, cmd, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    "%s: cannot add EPICS channel %s (hint: check PV name)",
                    self.name(),
                    cmd_name,
                )

        elif cmd_type.lower() == "sardana":
            doorname = None
            taurusname = None
            cmd_type = None
            door_first = False
            tango_first = False

            if "doorname" not in attributes_dict:
                try:
                    attributes_dict["doorname"] = self.doorname
                    doorname = self.doorname
                except AttributeError:
                    pass
            else:
                door_first = True
                doorname = attributes_dict["doorname"]

            if "taurusname" not in attributes_dict:
                try:
                    attributes_dict["taurusname"] = self.taurusname
                    taurusname = self.taurusname
                except AttributeError:
                    pass
            else:
                tango_first = True
                taurusname = attributes_dict["taurusname"]

            if "cmd_type" in attributes_dict:
                cmd_type = attributes_dict["cmd_type"]

            # guess what kind of command to create
            if cmd_type is None:
                if taurusname is not None and doorname is None:
                    cmd_type = "command"
                elif doorname is not None and taurusname is None:
                    cmd_type = "macro"
                elif doorname is not None and taurusname is not None:
                    if door_first:
                        cmd_type = "macro"
                    elif tango_first:
                        cmd_type = "command"
                    else:
                        cmd_type = "macro"
                else:
                    logging.getLogger().error(
                        "%s: incomplete sardana command declaration. ignored",
                        self.name(),
                    )

            from mxcubecore.Command.Sardana import (
                SardanaCommand,
                SardanaMacro,
            )

            if cmd_type == "macro" and doorname is not None:
                try:
                    new_command = SardanaMacro(cmd_name, cmd, **attributes_dict)
                except ConnectionError:
                    logging.getLogger().error(
                        "%s: could not connect to sardana door %s (hint: is it running ?)",
                        self.name(),
                        attributes_dict["doorname"],
                    )
                    raise ConnectionError
                except Exception:
                    logging.getLogger().exception(
                        '%s: could not add command "%s" (hint: check command attributes)',
                        self.name(),
                        cmd_name,
                    )
            elif cmd_type == "command" and taurusname is not None:
                try:
                    new_command = SardanaCommand(cmd_name, cmd, **attributes_dict)
                except ConnectionError:
                    logging.getLogger().error(
                        "%s: could not connect to sardana device %s (hint: is it running ?)",
                        self.name(),
                        taurusname,
                    )
                    raise ConnectionError
                except Exception:
                    logging.getLogger().exception(
                        '%s: could not add command "%s" (hint: check command attributes)',
                        self.name(),
                        cmd_name,
                    )
            else:
                logging.getLogger().error(
                    "%s: incomplete sardana command declaration. ignored", self.name()
                )

        elif cmd_type.lower() == "pool":
            if "tangoname" not in attributes_dict:
                try:
                    attributes_dict["tangoname"] = self.tangoname
                except AttributeError:
                    pass
            try:
                from mxcubecore.Command.Pool import PoolCommand

                new_command = PoolCommand(cmd_name, cmd, **attributes_dict)
            except ConnectionError:
                logging.getLogger().error(
                    "%s: could not connect to device server %s (hint: is it running ?)",
                    self.name(),
                    attributes_dict["tangoname"],
                )
                raise ConnectionError
            except Exception:
                logging.getLogger().exception(
                    '%s: could not add command "%s" (hint: check command attributes)',
                    self.name(),
                    cmd_name,
                )
        elif cmd_type.lower() == "tine":
            if "tinename" not in attributes_dict:
                try:
                    attributes_dict["tinename"] = self.tine_name
                except AttributeError:
                    pass

            try:
                from mxcubecore.Command.Tine import TineCommand

                new_command = TineCommand(cmd_name, cmd, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    '%s: could not add command "%s" (hint: check command attributes)',
                    self.name(),
                    cmd_name,
                )

        elif cmd_type.lower() == "mockup":
            try:
                from mxcubecore.Command.Mockup import MockupCommand

                new_command = MockupCommand(cmd_name, cmd, **attributes_dict)
            except Exception:
                logging.getLogger().exception(
                    '%s: could not add command "%s" (hint: check command attributes)',
                    self.name(),
                    cmd_name,
                )

        if new_command is not None:
            self.__commands[cmd_name] = new_command

            if not isinstance(arg1, dict):
                i = 1
                for arg in arg1.get_objects("argument"):
                    on_change = arg.get_property("onchange")
                    if on_change is not None:
                        on_change = (on_change, weakref.ref(self))
                    value_from = arg.get_property("valuefrom")
                    if value_from is not None:
                        value_from = (value_from, weakref.ref(self))

                    try:
                        combo_items = arg["type"]["item"]
                    except IndexError:
                        try:
                            new_command.add_argument(
                                arg.get_property("name"),
                                arg.type,
                                onchange=on_change,
                                valuefrom=value_from,
                            )
                        except AttributeError:
                            logging.getLogger().error(
                                '%s, command "%s": could not add argument %d, missing type or name',
                                self.name(),
                                cmd_name,
                                i,
                            )
                            continue
                    else:
                        if isinstance(combo_items, list):
                            combo_items = []
                            for item in combo_items:
                                name = item.get_property("name")
                                value = item.get_property("value")
                                if name is None or value is None:
                                    logging.getLogger().error(
                                        "%s, command '%s': could not add argument %d, missing combo item name or value",
                                        self.name(),
                                        cmd_name,
                                        i,
                                    )
                                    continue
                                else:
                                    combo_items.append((name, value))
                        else:
                            name = combo_items.get_property("name")
                            value = combo_items.get_property("value")
                            if name is None or value is None:
                                combo_items = ((name, value),)
                            else:
                                logging.getLogger().error(
                                    "%s, command '%s': could not add argument %d, missing combo item name or value",
                                    self.name(),
                                    cmd_name,
                                    i,
                                )
                                continue

                        new_command.add_argument(
                            arg.get_property("name"),
                            "combo",
                            combo_items,
                            on_change,
                            value_from,
                        )

                    i += 1

            return new_command

    def _add_channels_and_commands(self) -> None:
        """Add pending channels and commands."""
        [self.add_channel(*args) for args in self.__channels_to_add]
        [self.add_command(*args) for args in self.__commands_to_add]
        self.__channels_to_add = []
        self.__commands_to_add = []

    def execute_command(self, command_name: str, *args, **kwargs) -> Any:
        """Execute command.

        Args:
            command_name (str): Command name.
            *args (tuple): Arguments to pass through to the command to be executed.
            **kwargs (Dict[str, Any]): Named arguments to pass through to the
            command to be executed.

        Raises:
            AttributeError: If command not found.

        Returns:
            Any: Execution output.
        """
        if command_name in self.__commands:
            return self.__commands[command_name](*args, **kwargs)
        else:
            raise AttributeError



import logging
import math
import os
import time
from http.client import HTTPConnection
from urllib.parse import urlencode

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractMultiCollect import *
from mxcubecore.model.queue_model_objects import PathTemplate
from mxcubecore.utils.conversion import string_types


class ESRFMultiCollect(AbstractMultiCollect, HardwareObject):
    def __init__(self, name):
        AbstractMultiCollect.__init__(self)
        HardwareObject.__init__(self, name)
        self._centring_status = None
        self._metadataClient = None
        self.__mesh_steps = None
        self._mesh_range = None

    @property
    def _mesh_steps(self):
        return self.__mesh_steps

    @_mesh_steps.setter
    def _mesh_steps(self, steps):
        self.__mesh_steps = steps
        self._detector._mesh_steps = steps

    def execute_command(self, command_name, *args, **kwargs):
        wait = kwargs.get("wait", True)
        cmd_obj = self.get_command_object(command_name)
        return cmd_obj(*args, wait=wait)

    def init(self):
        self._detector = HWR.beamline.detector
        self.number_of_snapshots = self.get_property("num_snapshots", 4)

        self.setControlObjects(
            diffractometer=self.get_object_by_role("diffractometer"),
            sample_changer=self.get_object_by_role("sample_changer"),
            lims=self.get_object_by_role("dbserver"),
            safety_shutter=self.get_object_by_role("safety_shutter"),
            machine_current=self.get_object_by_role("machine_current"),
            cryo_stream=self.get_object_by_role("cryo_stream"),
            energy=self.get_object_by_role("energy"),
            resolution=self.get_object_by_role("resolution"),
            detector_distance=self.get_object_by_role("detector_distance"),
            transmission=self.get_object_by_role("transmission"),
            undulators=self.get_object_by_role("undulators"),
            flux=self.get_object_by_role("flux"),
            detector=self.get_object_by_role("detector"),
            beam_info=self.get_object_by_role("beam_info"),
        )

        try:
            undulators = self["undulator"]
        except KeyError:
            undulators = []

        beam_div_hor, beam_div_ver = HWR.beamline.beam.get_beam_divergence()

        self.setBeamlineConfiguration(
            synchrotron_name="ESRF",
            directory_prefix=self.get_property("directory_prefix"),
            default_exposure_time=HWR.beamline.detector.get_property(
                "default_exposure_time"
            ),
            minimum_exposure_time=HWR.beamline.detector.get_property(
                "minimum_exposure_time"
            ),
            detector_fileext=HWR.beamline.detector.get_property("file_suffix"),
            detector_type=HWR.beamline.detector.get_property("type"),
            detector_manufacturer=HWR.beamline.detector.get_property("manufacturer"),
            detector_model=HWR.beamline.detector.get_property("model"),
            detector_binning_mode=HWR.beamline.detector.get_property("binning_mode"),
            detector_px=HWR.beamline.detector.get_property("px"),
            detector_py=HWR.beamline.detector.get_property("py"),
            undulators=undulators,
            focusing_optic=self.get_property("focusing_optic"),
            monochromator_type=self.get_property("monochromator"),
            beam_divergence_vertical=beam_div_ver,
            beam_divergence_horizontal=beam_div_hor,
            polarisation=self.get_property("polarisation"),
            maximum_phi_speed=self.get_property("maximum_phi_speed"),
            minimum_phi_oscillation=self.get_property("minimum_phi_oscillation"),
            input_files_server=self.get_property("input_files_server"),
        )

        # self._detector.init(HWR.beamline.detector, self)

        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    @task
    def data_collection_end_hook(self, data_collect_parameters):
        self._detector._emit_status()

    def prepare_oscillation(
        self,
        start,
        osc_range,
        exptime,
        number_of_images,
        shutterless,
        first_frame,
    ):
        if shutterless:
            end = start + osc_range * number_of_images
            exptime = (exptime + self._detector.get_deadtime()) * number_of_images

            if first_frame:
                self.do_prepare_oscillation(start, end, exptime)
        else:
            if osc_range < 1e-4:
                # still image
                end = start
            else:
                end = start + osc_range
                self.do_prepare_oscillation(start, end, exptime)

        return start, end

    @task
    def no_oscillation(self, exptime):
        self.open_fast_shutter()
        time.sleep(exptime)
        self.close_fast_shutter()


    @task
    def do_oscillation(
        self, start, end, exptime, number_of_images, shutterless, first_frame,
    ):
        if shutterless:
            if first_frame:
                exptime = (exptime + self._detector.get_deadtime()) * number_of_images
                self.oscillation_task = self.oscil(
                    start, end, exptime, number_of_images, wait=False
                )

            if self.oscillation_task.ready():
                self.oscillation_task.get()
        else:
            self.oscil(start, end, exptime, number_of_images)

    @task
    def oscil(self, start, end, exptime, number_of_images, wait=False):
        if math.fabs(end - start) < 1e-4:
            self.open_fast_shutter()
            time.sleep(exptime)
            self.close_fast_shutter()
        else:
            return self.execute_command("do_oscillation", start, end, exptime, number_of_images)

    def set_wavelength(self, wavelength):
        if HWR.beamline.tunable_wavelength:
            return HWR.beamline.energy.set_wavelength(wavelength)
        else:
            return

    def set_energy(self, energy):
        if HWR.beamline.tunable_wavelength:
            return HWR.beamline.energy.set_value(energy)
        else:
            return

    def data_collection_cleanup(self):
        try:
            self.close_fast_shutter()
            self.stop_oscillation()
            HWR.beamline.detector.stop_acquisition()
        except Exception:
            logging.getLogger("HWR").exception("")

    def queue_finished_cleanup(self):
        logging.getLogger("user_level_log").info("Queue execution finished")

    @task
    def close_fast_shutter(self):
        self.execute_command("close_fast_shutter")

    @task
    def open_fast_shutter(self):
        self.execute_command("open_fast_shutter")

    @task
    def move_motors(self, motor_position_dict):
        # We do not wnta to modify the input dict
        motor_positions_copy = motor_position_dict.copy()
        for motor in motor_positions_copy.keys():  # iteritems():
            position = motor_positions_copy[motor]
            if isinstance(motor, string_types):
                # find right motor object from motor role in diffractometer obj.
                motor_role = motor
                motor = HWR.beamline.diffractometer.get_deviceby_role(motor_role)
                del motor_positions_copy[motor_role]
                if motor is None:
                    continue
                motor_positions_copy[motor] = position

            logging.getLogger("HWR").info(
                "Moving motor '%s' to %f", motor.get_motor_mnemonic(), position
            )
            motor.set_value(position)

        while any([motor.motorIsMoving() for motor in motor_positions_copy]):
            logging.getLogger("HWR").info("Waiting for end of motors motion")
            time.sleep(0.02)

    def open_safety_shutter(self):
        if HWR.beamline.safety_shutter.get_value().name == "CLOSED":
            try:
                logging.getLogger("user_level_log").info("Opening safety shutter")
                HWR.beamline.safety_shutter.set_value(
                    HWR.beamline.safety_shutter.VALUES.OPEN, timeout=10
                )
            except Exception:
                logging.getLogger("HWR").exception("")

    def safety_shutter_opened(self):
        state = False

        try:
            state = HWR.beamline.safety_shutter.get_value().name == "OPEN"
        except Exception:
            logging.getLogger("HWR").exception("")
            state = True

        return state

    @task
    def close_safety_shutter(self):
        if HWR.beamline.safety_shutter.get_value().name == "OPEN":
            try:
                HWR.beamline.safety_shutter.set_value(
                    HWR.beamline.safety_shutter.VALUES.CLOSED
                )
            except Exception:
                logging.getLogger("HWR").exception("")

    @task
    def prepare_intensity_monitors(self):
        try:
            self.execute_command("adjust_gains")
        except AttributeError:
            pass

    def prepare_acquisition(
        self,
        take_dark,
        start,
        osc_range,
        exptime,
        npass,
        number_of_images,
        comment="",
    ):
        self._detector.prepare_acquisition(
            take_dark,
            start,
            osc_range,
            exptime,
            npass,
            number_of_images,
            comment,
            self.mesh,
            self.mesh_num_lines,
        )

    def set_detector_filenames(
        self, is_first_frame, frame_number, start, filename, shutterless
    ):
        if is_first_frame or not shutterless:
            return self._detector.set_detector_filenames(frame_number, start, filename)

    def stop_oscillation(self):
        HWR.beamline.diffractometer.abort_cmd()

    def start_acquisition(self, exptime, npass, first_frame, shutterless):
        if first_frame:
            return self._detector.start_acquisition()

    @task
    def write_image(self, last_frame):
        if last_frame:
            res = self._detector.wait_ready()
            self._detector._emit_status()
            return res

    def last_image_saved(self, total_time, exptime, num_images):
        # check here if fast shutter is open ?
        if HWR.beamline.detector.status["acq_satus"] == "RUNNING":
            return int(total_time / exptime)
        else:
            return HWR.beamline.detector.last_image_saved()

    def stop_acquisition(self):
        return self._detector.stop_acquisition()

    def reset_detector(self):
        return self._detector.reset()

    def prepare_input_files(
        self, files_directory, prefix, run_number, process_directory
    ):

        autoprocessing_input_file_dirname = "autoprocessing"
        autoprocessing_directory = os.path.join(
            process_directory, autoprocessing_input_file_dirname
        )
        xds_input_file_dirname = "xds_%s_run%s" % (prefix, run_number,)
        mosflm_input_file_dirname = "mosflm_%s_run%s" % (prefix, run_number)
        hkl2000_dirname = "hkl2000_%s_run%s" % (prefix, run_number)

        i = 1
        while os.path.exists(autoprocessing_directory):
            autoprocessing_input_file_dirname = "autoprocessing_%d" % i
            autoprocessing_directory = os.path.join(
                process_directory, autoprocessing_input_file_dirname
            )
            xds_input_file_dirname = "xds_%s_run%s_%d" % (prefix, run_number, i)
            mosflm_input_file_dirname = "mosflm_%s_run%s_%d" % (prefix, run_number, i)
            hkl2000_dirname = "hkl2000_%s_run%s_%d" % (prefix, run_number, i)
            i += 1

        self.raw_data_input_file_dir = os.path.join(
            files_directory, "process", xds_input_file_dirname
        )
        self.mosflm_raw_data_input_file_dir = os.path.join(
            files_directory, "process", mosflm_input_file_dirname
        )
        self.raw_hkl2000_dir = os.path.join(files_directory, "process", hkl2000_dirname)

        for dir0 in (
            self.raw_data_input_file_dir,
            self.mosflm_raw_data_input_file_dir,
            self.raw_hkl2000_dir,
            autoprocessing_directory,
        ):
            self.create_directories(dir0)
            logging.info("Creating processing input file directory: %s", dir0)
            os.chmod(dir0, 0o777)

        try:
            try:
                os.symlink(files_directory, os.path.join(process_directory, "links"))
            except os.error as e:
                if e.errno != errno.EEXIST:
                    raise
        except Exception:
            logging.exception("Could not create processing file directory")

        return autoprocessing_directory, "", ""

    @task
    def write_input_files(self, collection_id):
        # assumes self.xds_directory and self.mosflm_directory are valid
        conn = HTTPConnection(self.bl_config.input_files_server)

        # hkl input files
        input_file_dir = self.raw_hkl2000_dir
        file_prefix = "../.."
        hkl_file_path = os.path.join(input_file_dir, "def.site")

        conn.request("GET", "/def.site/%d?basedir=%s" % (collection_id, file_prefix))
        hkl_file = open(hkl_file_path, "w")
        r = conn.getresponse()

        if r.status != 200:
            logging.error("Could not create hkl input file")
        else:
            hkl_file.write(r.read().decode())
        hkl_file.close()
        os.chmod(hkl_file_path, 0o666)

        for input_file_dir, file_prefix in (
            (self.raw_data_input_file_dir, "../.."),
            (self.xds_directory, "../links"),
        ):
            xds_input_file = os.path.join(input_file_dir, "XDS.INP")
            conn.request("GET", "/xds.inp/%d?basedir=%s" % (collection_id, file_prefix))
            xds_file = open(xds_input_file, "w")
            res = conn.getresponse()
            if res.status != 200:
                logging.error("Could not create xds input file")
            else:
                xds_file.write(res.read().decode())
            xds_file.close()
            os.chmod(xds_input_file, 0o666)

        input_file_dir = self.mosflm_raw_data_input_file_dir
        file_prefix = "../.."
        mosflm_input_file = os.path.join(input_file_dir, "mosflm.inp")
        conn.request("GET", "/mosflm.inp/%d?basedir=%s" % (collection_id, file_prefix))
        mosflm_file = open(mosflm_input_file, "w")
        mosflm_file.write(conn.getresponse().read().decode())
        mosflm_file.close()
        os.chmod(mosflm_input_file, 0o666)

        # also write input file for STAC
        for stac_om_input_file_name, stac_om_dir in (
            ("xds.descr", self.xds_directory),
            ("mosflm.descr", self.mosflm_raw_data_input_file_dir),
            ("xds.descr", self.raw_data_input_file_dir),
        ):
            stac_om_input_file = os.path.join(stac_om_dir, stac_om_input_file_name)
            conn.request("GET", "/stac.descr/%d" % collection_id)
            stac_om_file = open(stac_om_input_file, "w")
            stac_template = conn.getresponse().read().decode()
            if stac_om_input_file_name.startswith("xds"):
                om_type = "xds"
                if stac_om_dir == self.raw_data_input_file_dir:
                    om_filename = os.path.join(stac_om_dir, "CORRECT.LP")
                else:
                    om_filename = os.path.join(
                        stac_om_dir, "xds_fastproc", "CORRECT.LP"
                    )
            else:
                om_type = "mosflm"
                om_filename = os.path.join(stac_om_dir, "bestfile.par")

            stac_om_file.write(
                stac_template.format(
                    omfilename=om_filename,
                    omtype=om_type,
                    phi=HWR.beamline.diffractometer.phiMotor.get_value(),
                    sampx=HWR.beamline.diffractometer.sampleXMotor.get_value(),
                    sampy=HWR.beamline.diffractometer.sampleYMotor.get_value(),
                    phiy=HWR.beamline.diffractometer.phiyMotor.get_value(),
                )
            )
            stac_om_file.close()
            os.chmod(stac_om_input_file, 0o666)

    def get_wavelength(self):
        return HWR.beamline.energy.get_wavelength()

    def get_undulators_gaps(self):
        all_gaps = {"Unknown": None}
        _gaps = {}

        try:
            _gaps = HWR.beamline.undulators
        except Exception:
            logging.getLogger("HWR").exception("Could not get undulator gaps")
        all_gaps.clear()
        for key in _gaps:
            if "_Position" in key:
                nkey = key[:-9]
                all_gaps[nkey] = _gaps[key]
            else:
                all_gaps = _gaps
        return all_gaps

    def get_resolution_at_corner(self):
        return self.execute_command("get_resolution_at_corner")

    def get_beam_size(self):
        return (
            self.execute_command("get_beam_size_x"),
            self.execute_command("get_beam_size_y"),
        )

    def get_slit_gaps(self):
        return (
            self.execute_command("get_slit_gap_h"),
            self.execute_command("get_slit_gap_v"),
        )

    def get_beam_shape(self):
        return self.execute_command("get_beam_shape")

    # def get_measured_intensity(self):
    #     try:
    #         val = self.get_channel_object("image_intensity").getValue()
    #         return float(val)
    #     except Exception:
    #         return 0

    def get_machine_current(self):
        if HWR.beamline.machine_info:
            try:
                return HWR.beamline.machine_info.get_current()
            except Exception:
                return -1
        else:
            return 0

    def get_machine_message(self):
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_message()
        else:
            return ""

    def get_machine_fill_mode(self):
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_fill_mode()
        else:
            """"""

    def get_cryo_temperature(self):
        while True:
            logging.info("Reading cryostream temperature")
            try:
                T = HWR.beamline.cryo_stream.getTemperature()
            except Exception:
                time.sleep(0.1)
                continue
            else:
                return T

    def get_current_energy(self):
        return HWR.beamline.energy.get_value()

    def get_beam_centre(self):
        return (
            self.execute_command("get_beam_centre_x"),
            self.execute_command("get_beam_centre_y"),
        )

    def getBeamlineConfiguration(self, *args):
        # TODO: change this to stop using a dictionary at the other end
        return self.bl_config._asdict()

    def is_connected(self):
        return True

    def is_ready(self):
        return True

    def sampleChangerHO(self):
        return HWR.beamline.sample_changer

    def diffractometer(self):
        return HWR.beamline.diffractometer

    def dbServerHO(self):
        return HWR.beamline.lims

    def sanityCheck(self, collect_params):
        return

    def setBrick(self, brick):
        return

    def directoryPrefix(self):
        return self.bl_config.directory_prefix

    def store_image_in_lims(self, frame, first_frame, last_frame):
        if first_frame or last_frame:
            return True

    @task
    def generate_image_jpeg(self, filename, jpeg_path, jpeg_thumbnail_path):
        directories = filename.split(os.path.sep)
        try:
            if directories[1] == "data" and directories[2] == "gz":
                if directories[3] == "visitor":
                    proposal = directories[4]
                    beamline = directories[5]
                elif directories[4] == "inhouse":
                    proposal = directories[5]
                    beamline = directories[3]
                else:
                    proposal = "unknown"
                    beamline = "unknown"

            elif directories[2] == "visitor":
                beamline = directories[4]
                proposal = directories[3]
            else:
                beamline = directories[2]
                proposal = directories[4]
        except Exception:
            beamline = "unknown"
            proposal = "unknown"
        host, port = self.get_property("bes_jpeg_hostport").split(":")
        conn = HTTPConnection(host, int(port))

        params = urlencode(
            {
                "image_path": filename,
                "jpeg_path": jpeg_path,
                "jpeg_thumbnail_path": jpeg_thumbnail_path,
                "initiator": beamline,
                "externalRef": proposal,
                "reuseCase": "true",
            }
        )
        conn.request(
            "POST",
            "/BES/bridge/rest/processes/CreateThumbnails/RUN?%s" % params,
            headers={"Accept": "text/plain"},
        )
        conn.getresponse()

    """
    getOscillation
        Description: Returns the parameters (and results) of an oscillation.
        Type       : method
        Arguments  : oscillation_id (int; the oscillation id, the last parameters of the collectOscillationStarted
                                     signal)
        Returns    : tuple; (blsampleid,barcode,location,parameters)
    """

    def getOscillation(self, oscillation_id):
        return self.oscillations_history[oscillation_id - 1]

    def sampleAcceptCentring(self, accepted, centring_status):
        self.sample_centring_done(accepted, centring_status)

    def setCentringStatus(self, centring_status):
        self._centring_status = centring_status

    """
    getOscillations
        Description: Returns the history of oscillations for a session
        Type       : method
        Arguments  : session_id (int; the session id, stored in the "sessionId" key in each element
                                 of the parameters list in the collect method)
        Returns    : list; list of all oscillation_id for the specified session
    """

    def getOscillations(self, session_id):
        # TODO
        return []

    def set_helical(self, helical_on):
        self.get_channel_object("helical").set_value(1 if helical_on else 0)

    def set_helical_pos(self, helical_oscil_pos):
        self.get_channel_object("helical_pos").set_value(helical_oscil_pos)

    def get_archive_directory(self, directory):
        pt = PathTemplate()
        pt.directory = directory
        return pt.get_archive_directory()

    def setMeshScanParameters(self, mesh_steps, mesh_range):
        """
        Descript. :
        """
        self._mesh_steps = mesh_steps
        self._mesh_range = mesh_range



import logging
import sys

import gevent
import PyTango.gevent

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR

"""
Read the state of the hutch from the PSS device server and take actions
when enter (1) or interlock (0) the hutch.
0 = The hutch has been interlocked and the sample environment should be made
     ready for either loading a sample or a data collection.
1 = The interlock is cleared and the user is entering the hutch. The equipment
    is placed in safe position.
Example xml file:
<object class="ESRF.BlissHutchTrigger">
  <username>Hutchtrigger</username>
  <pss_tangoname>orion:10000/exp/id29-cr1/sl0</pss_tangoname>
  <pss_card_ch>7/1</pss_card_ch>
  <polling_interval>500</polling_interval>
  <object href="/bliss" role="controller"/>
  <object href="/sample_changer", role="sample_changer"/>
</object>
"""


class BlissHutchTrigger(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        BaseHardwareObjects.HardwareObject.__init__(self, name)
        self._enabled = True

    def _do_polling(self):
        while True:
            try:
                self.poll()
            except Exception:
                sys.excepthook(*sys.exc_info())
            gevent.sleep(self.get_property("polling_interval") / 1000.0 or 1)

    def init(self):
        try:
            self.device = PyTango.gevent.DeviceProxy(self.get_property("pss_tangoname"))
        except PyTango.DevFailed as traceback:
            last_error = traceback[-1]
            logging.getLogger("HWR").error(
                "%s: %s", str(self.name()), last_error["desc"]
            )
            self.device = None

        self.pollingTask = None
        self.initialized = False
        self.__oldValue = None
        self.card = None
        self.channel = None

        PSSinfo = self.get_property("pss_card_ch")
        try:
            self.card, self.channel = map(int, PSSinfo.split("/"))
        except Exception:
            logging.getLogger().error("%s: cannot find PSS number", self.name())
            return

        if self.device is not None:
            self.pollingTask = gevent.spawn(self._do_polling)
        self.connected()

    def hutchIsOpened(self):
        return self.hutch_opened

    def is_connected(self):
        return True

    def connected(self):
        self.emit("connected")

    def disconnected(self):
        self.emit("disconnected")

    def abort(self):
        pass

    def macro(self, entering_hutch, **kwargs):
        logging.info(
            "%s: %s hutch", self.name(), "entering" if entering_hutch else "leaving"
        )
        ctrl_obj = self.get_object_by_role("controller")
        ctrl_obj.hutch_actions(entering_hutch, hutch_trigger=True, **kwargs)

        # open the flexHCD ports
        sample_changer = HWR.beamline.sample_changer
        if sample_changer:
            if entering_hutch:
                sample_changer.prepare_hutch(robot_port=0)
            else:
                sample_changer.prepare_hutch(user_port=0, robot_port=1)

    def poll(self):
        a = self.device.GetInterlockState([self.card - 1, 2 * (self.channel - 1)])[0]
        b = self.device.GetInterlockState([self.card - 1, 2 * (self.channel - 1) + 1])[
            0
        ]
        value = a & b

        if value == self.__oldValue:
            return
        else:
            self.__oldValue = value

        self.value_changed(value)

    def value_changed(self, value, *args):
        if value == 0:
            if self.initialized:
                self.emit("hutchTrigger", (1,))
        elif value == 1 and self.initialized:
            self.emit("hutchTrigger", (0,))

        self.hutch_opened = 1 - value
        self.initialized = True
        if self._enabled:
            self.macro(self.hutch_opened)

    def get_actuator_state(self):
        if self._enabled:
            return "ENABLED"
        else:
            return "DISABLED"

    def actuatorIn(self):
        self._enabled = True

    def actuatorOut(self):
        self._enabled = False



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.
"""Abstract machine info class"""

import abc
from ast import literal_eval

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractMachineInfo(HardwareObject):
    """Abstract machine info - information coming from the accelerator source.
       It provides only few common to all accelerators parameters.

    Emits:
        valueChanged: ("valueChanged", (value,))

    Attributes:
        _mach_info_dict: Dictionary with all the defined parameters.
        _mach_info_keys: List of keys to be present in the above dictionary.

    Note:
        The methods to be used to fill the _mach_info_dict should start with
        get_ (e.g. get_current for reading the current)
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self._mach_info_dict = {}
        self._mach_info_keys = []

    def init(self):
        """Get the attributes to be defined as keys in the _machine_info_dict"""
        attr = self.get_property("parameters")
        if attr:
            self._check_attributes(literal_eval(attr))
        else:
            # at least current should be in the list
            self._check_attributes(["current"])

    @abc.abstractmethod
    def get_current(self) -> float:
        """Read the ring current.

        Returns:
            Current [mA].
        """
        return 0

    def get_message(self) -> str:
        """Read the operator's message.

        Returns:
            Message.
        """
        return ""

    def get_lifetime(self) -> float:
        """Read the lifetime.

        Returns:
            Lifetime [s].
        """
        return 0

    def get_topup_remaining(self) -> float:
        """Read the top-up remaining time.

        Returns:
            Top-up remaining [s].
        """
        return 0

    def get_fill_mode(self) -> str:
        """Read the fill mode as text.

        Returns:
            Machine fill mode
        """
        return ""

    def get_value(self) -> dict:
        """Read machine info summary as dictionary.

        Returns:
            Copy of the _mach_info_dict.
        """
        for val in dir(self):
            if val.startswith("get_"):
                if val[4:] in self._mach_info_keys:
                    self._mach_info_dict.update({val[4:]: getattr(self, val)()})
        return self._mach_info_dict.copy()

    def _check_attributes(self, attr_list=None):
        """Check if all the keys in the configuration file have
        implemented read method. Remove the undefined.
        """
        attr_list = attr_list or self._mach_info_keys

        for attr_key in attr_list:
            try:
                getattr(self, f"get_{attr_key}")
            except AttributeError:
                attr_list.remove(attr_key)
        self._mach_info_keys = attr_list

    def update_value(self, value=None):
        """Emits signal valueChanged.

        Args:
            value(dict): Dictionary with all the value
        """
        if value is None:
            value = self.get_value()

        self.emit("valueChanged", (value,))



import logging
import os
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractMultiCollect import (
    AbstractMultiCollect,
)
from mxcubecore.TaskUtils import task


class LNLSCollect(AbstractMultiCollect, HardwareObject):
    def __init__(self, name):
        AbstractMultiCollect.__init__(self)
        HardwareObject.__init__(self, name)
        self._centring_status = None
        self.ready_event = None
        self.actual_frame_num = 0

    def execute_command(self, command_name, *args, **kwargs):
        return

    def init(self):
        self.setControlObjects(
            diffractometer=self.get_object_by_role("diffractometer"),
            sample_changer=self.get_object_by_role("sample_changer"),
            lims=self.get_object_by_role("dbserver"),
            safety_shutter=self.get_object_by_role("safety_shutter"),
            machine_current=self.get_object_by_role("machine_current"),
            cryo_stream=self.get_object_by_role("cryo_stream"),
            energy=HWR.beamline.energy,
            resolution=self.get_object_by_role("resolution"),
            detector_distance=self.get_object_by_role("detector_distance"),
            transmission=self.get_object_by_role(
                "transmission"
            ),  # Returns attenuators.
            undulators=self.get_object_by_role("undulators"),
            flux=self.get_object_by_role("flux"),
            detector=self.get_object_by_role("detector"),
            beam_info=self.get_object_by_role("beam_info"),
        )
        # Adding this line to get transmission value:
        self.filter_transmission = self.get_object_by_role("filter_transmission")
        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    @task
    def loop(self, owner, data_collect_parameters_list):
        print("\nCALL LOOP\n")
        print("\nDC PARAM LIST:\n")
        print(
            "data_collect_parameters_list = {}\n".format(data_collect_parameters_list)
        )
        failed_msg = "Data collection failed!"
        failed = True
        collections_analyse_params = []
        self.emit("collectReady", (False,))
        self.emit("collectStarted", (owner, 1))

        for data_collect_parameters in data_collect_parameters_list:
            print("\nSUB LOOP SINGLE COLLECT?\n")
            print("data_collect_parameters = {}\n".format(data_collect_parameters))
            logging.debug("collect parameters = %r", data_collect_parameters)
            failed = False
            data_collect_parameters["status"] = "Data collection successful"
            (
                osc_id,
                sample_id,
                sample_code,
                sample_location,
            ) = self.update_oscillations_history(data_collect_parameters)

            self.emit(
                "collectOscillationStarted",
                (
                    owner,
                    sample_id,
                    sample_code,
                    sample_location,
                    data_collect_parameters,
                    osc_id,
                ),
            )

            # Translate parameters to scan-utils flyscan
            config_yml = "pilatus"
            message = "Flyscan called from mxcubeweb."

            output_directory = data_collect_parameters["fileinfo"]["directory"]
            # Create dir
            path = output_directory
            try:
                if os.path.isdir(path):
                    logging.getLogger("HWR").info("Directory exists: %s " % path)
                else:
                    os.makedirs(path)
                    logging.getLogger("HWR").info(
                        "Successfully created the directory %s " % path
                    )
            except OSError:
                logging.getLogger("HWR").error(
                    "Creation of the directory %s failed." % path
                )

            if not output_directory.endswith("/"):
                output_directory = output_directory + "/"
            output_prefix = data_collect_parameters["fileinfo"]["prefix"]
            output_file = output_directory + output_prefix

            motor_mnenomic = "gonio"
            xlabel = motor_mnenomic
            plot_type = "none"
            mode = "--points-mode"

            start_float = float(
                data_collect_parameters["oscillation_sequence"][0]["start"]
            )  # omega start pos
            start = str(start_float)

            step_size = float(
                data_collect_parameters["oscillation_sequence"][0]["range"]
            )
            num_of_points = int(
                data_collect_parameters["oscillation_sequence"][0]["number_of_images"]
            )
            end_float = start_float + step_size * num_of_points
            end = str(end_float)

            step_or_points = str(num_of_points)

            time_float = float(
                data_collect_parameters["oscillation_sequence"][0]["exposure_time"]
            )
            time = str(time_float)

            prescan = " "
            postscan = " "

            # flyscan-only params
            start_offset = str(0)
            end_offset = str(0)
            aquire_period = str(time_float + 0.0023)  # + pilatus readout time

            command = 'flyscan -c {} -m "{}" -o {} -s --motor {} {} --start {} --end {} --step-or-points {} --time {} --prescan={} --postscan={} --start-offset {} --end-offset {} --aquire-period {}'.format(
                config_yml,
                message,
                output_file,
                motor_mnenomic,
                mode,
                start,
                end,
                step_or_points,
                time,
                prescan,
                postscan,
                start_offset,
                end_offset,
                aquire_period,
            )

            # command = 'scan -c {} -m "{}" -o {} --motor {} --xlabel {} --plot-type {} {} --start {} --end {} --step-or-points {} --time {} --prescan={} --postscan={}'.format(config_yml, message, output_file, motor_mnenomic, xlabel, plot_type, mode, start, end, step_or_points, time, prescan, postscan)

            logging.getLogger("HWR").info("[SCAN-UTILS] Command: " + str(command))
            print("\n[SCAN-UTILS] Command: " + str(command) + "\n")

            # Store values for clean up
            logging.getLogger("HWR").info("[Clean up] Configuring...")
            omega = HWR.beamline.diffractometer.motor_hwobj_dict.get("phi")
            if omega is None:
                logging.getLogger("HWR").error("[Clean up] Could not get omega motor.")
            else:
                omega_original_velo = omega.get_velocity()
                logging.getLogger("HWR").info(
                    "[Clean up] Omega velo: {}".format(omega_original_velo)
                )
            logging.getLogger("HWR").info("[Clean up] Configured.")

            # Set detector cbf header
            header_ok = self.set_pilatus_det_header(start_float, step_size)
            if not header_ok:
                logging.getLogger("HWR").error(
                    "[Collect] Pilatus header params could not be set! Collection aborted."
                )
                return

            import subprocess
            import sys

            try:
                process = subprocess.Popen(
                    command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
                )

                logging.getLogger("HWR").info("[SCAN-UTILS] Executing scan...")
                logging.getLogger("user_level_log").info("Executing scan...")
                stdout, stderr = process.communicate()
                # stdout
                logging.getLogger("HWR").info(
                    "[SCAN-UTILS] output : " + stdout.decode("utf-8")
                )
                print("[SCAN-UTILS] output : " + stdout.decode("utf-8"))
                # stderr
                logging.getLogger("HWR").error(
                    "[SCAN-UTILS] errors : " + stderr.decode("utf-8")
                )
                print("[SCAN-UTILS] errors : " + stderr.decode("utf-8"))

            except BaseException:
                logging.getLogger("HWR").error("[SCAN-UTILS] Error in calling scan.")
                # print("[SCAN-UTILS] Error in calling scan.")
                raise
            else:
                logging.getLogger("HWR").info("[SCAN-UTILS] Finished scan!")
                logging.getLogger("user_level_log").info("Finished scan!")
                # print("[SCAN-UTILS] Finished scan!")
            finally:
                # Clean up
                logging.getLogger("HWR").info("[Clean up] Applying...")
                import time as timee

                timee.sleep(10)
                if omega is not None:
                    # Restore omega default velocity
                    omega.set_velocity(omega_original_velo)
                    logging.getLogger("HWR").info(
                        "[Clean up] Omega velo reset to: {}".format(omega_original_velo)
                    )
                logging.getLogger("HWR").info("[Clean up] Done!")

            data_collect_parameters["status"] = "Running"
            data_collect_parameters["status"] = "Data collection successful"
            self.emit(
                "collectOscillationFinished",
                (
                    owner,
                    True,
                    data_collect_parameters["status"],
                    "12345",
                    osc_id,
                    data_collect_parameters,
                ),
            )

        self.emit(
            "collectEnded",
            owner,
            not failed,
            failed_msg if failed else "Data collection successful",
        )
        logging.getLogger("HWR").info("data collection successful in loop")
        self.emit("collectReady", (True,))

    def set_pilatus_det_header(self, start_angle, step_size):
        # Read current params
        logging.getLogger("HWR").info("Setting Pilatus CBF header...")
        wl = self.bl_control.energy.get_wavelength()
        dd = self.bl_control.detector_distance.get_value()
        te = self.bl_control.energy.get_value()
        ft = self.filter_transmission.get_value()
        try:
            ft = ft / 100  # [0, 1]
        except Exception as e:
            print("Error on setting Pilatus transmission: {}".format(str(e)))
            return False

        # Write to det (values will be on the cbf header)
        wl_ok = self.bl_control.detector.set_wavelength(wl)
        dd_ok = self.bl_control.detector.set_detector_distance(dd)
        bx_ok = self.bl_control.detector.set_beam_x(from_user=True)
        by_ok = self.bl_control.detector.set_beam_y(from_user=True)
        te_ok = self.bl_control.detector.set_threshold_energy(te)
        ft_ok = self.bl_control.detector.set_transmission(ft)
        sa_ok = self.bl_control.detector.set_start_angle(start_angle)
        ss_ok = self.bl_control.detector.set_angle_incr(step_size)

        return (
            wl_ok and dd_ok and bx_ok and by_ok and te_ok and ft_ok and sa_ok and ss_ok
        )

    @task
    def data_collection_hook(self, data_collect_parameters):
        return

    def do_prepare_oscillation(self, start, end, exptime, npass):
        self.actual_frame_num = 0

    @task
    def oscil(self, start, end, exptime, npass):
        return

    @task
    def data_collection_cleanup(self):
        return

    @task
    def close_fast_shutter(self):
        return

    @task
    def open_fast_shutter(self):
        return

    @task
    def move_motors(self, motor_position_dict):
        return

    @task
    def open_safety_shutter(self):
        return

    def safety_shutter_opened(self):
        return

    @task
    def close_safety_shutter(self):
        return

    @task
    def prepare_intensity_monitors(self):
        return

    def prepare_acquisition(
        self, take_dark, start, osc_range, exptime, npass, number_of_images, comment=""
    ):
        return

    def set_detector_filenames(
        self, frame_number, start, filename, jpeg_full_path, jpeg_thumbnail_full_path
    ):
        return

    def prepare_oscillation(
        self, start, osc_range, exptime, number_of_images, shutterless, first_frame
    ):
        return (start, start + osc_range)

    def do_oscillation(
        self, start, end, exptime, number_of_images, shutterless, first_frame
    ):
        gevent.sleep(exptime)

    def start_acquisition(self, exptime, npass, first_frame):
        return

    def write_image(self, last_frame):
        self.actual_frame_num += 1
        return

    def last_image_saved(self):
        return self.actual_frame_num

    def stop_acquisition(self):
        return

    def reset_detector(self):
        return

    def prepare_input_files(
        self, files_directory, prefix, run_number, process_directory
    ):
        self.actual_frame_num = 0
        i = 1
        while True:
            xds_input_file_dirname = "xds_%s_run%s_%d" % (prefix, run_number, i)
            xds_directory = os.path.join(process_directory, xds_input_file_dirname)

            if not os.path.exists(xds_directory):
                break

            i += 1

        mosflm_input_file_dirname = "mosflm_%s_run%s_%d" % (prefix, run_number, i)
        mosflm_directory = os.path.join(process_directory, mosflm_input_file_dirname)

        hkl2000_dirname = "hkl2000_%s_run%s_%d" % (prefix, run_number, i)
        hkl2000_directory = os.path.join(process_directory, hkl2000_dirname)

        self.raw_data_input_file_dir = os.path.join(
            files_directory, "process", xds_input_file_dirname
        )
        self.mosflm_raw_data_input_file_dir = os.path.join(
            files_directory, "process", mosflm_input_file_dirname
        )
        self.raw_hkl2000_dir = os.path.join(files_directory, "process", hkl2000_dirname)

        return xds_directory, mosflm_directory, hkl2000_directory

    @task
    def write_input_files(self, collection_id):
        return

    def get_wavelength(self):
        return

    def get_undulators_gaps(self):
        return []

    def get_resolution_at_corner(self):
        return

    def get_beam_size(self):
        return None, None

    def get_slit_gaps(self):
        return None, None

    def get_beam_shape(self):
        return

    def get_machine_current(self):
        if self.bl_control.machine_current is not None:
            return self.bl_control.machine_current.getCurrent()
        else:
            return 0

    def get_machine_message(self):
        if self.bl_control.machine_current is not None:
            return self.bl_control.machine_current.getMessage()
        else:
            return ""

    def get_machine_fill_mode(self):
        if self.bl_control.machine_current is not None:
            return self.bl_control.machine_current.getFillMode()
        else:
            """"""

    def get_cryo_temperature(self):
        if self.bl_control.cryo_stream is not None:
            return self.bl_control.cryo_stream.getTemperature()

    def get_current_energy(self):
        return

    def get_beam_centre(self):
        return None, None

    def getBeamlineConfiguration(self, *args):
        return self.bl_config._asdict()

    def isConnected(self):
        return True

    def is_ready(self):
        return True

    def sampleChangerHO(self):
        return self.bl_control.sample_changer

    def diffractometer(self):
        return self.bl_control.diffractometer

    def sanityCheck(self, collect_params):
        return

    def setBrick(self, brick):
        return

    def directoryPrefix(self):
        return self.bl_config.directory_prefix

    def store_image_in_lims(self, frame, first_frame, last_frame):
        return True

    def getOscillation(self, oscillation_id):
        return self.oscillations_history[oscillation_id - 1]

    def sampleAcceptCentring(self, accepted, centring_status):
        self.sample_centring_done(accepted, centring_status)

    def setCentringStatus(self, centring_status):
        self._centring_status = centring_status

    def getOscillations(self, session_id):
        return []

    def set_helical(self, helical_on):
        return

    def set_helical_pos(self, helical_oscil_pos):
        return

    def get_archive_directory(self, directory):
        archive_dir = os.path.join(directory, "archive")
        return archive_dir

    @task
    def generate_image_jpeg(self, filename, jpeg_path, jpeg_thumbnail_path):
        pass



import logging
import time

from mxcubecore.HardwareObjects.abstract import AbstractSampleChanger
from mxcubecore.HardwareObjects.abstract.sample_changer import Container


class SampleChangerMockup(AbstractSampleChanger.SampleChanger):
    __TYPE__ = "Mockup"
    NO_OF_BASKETS = 5
    NO_OF_SAMPLES_IN_BASKET = 10

    def __init__(self, *args, **kwargs):
        super(SampleChangerMockup, self).__init__(self.__TYPE__, False, *args, **kwargs)

    def init(self):
        self._selected_sample = -1
        self._selected_basket = -1
        self._scIsCharging = None

        self.no_of_baskets = self.get_property(
            "no_of_baskets", SampleChangerMockup.NO_OF_BASKETS
        )

        self.no_of_samples_in_basket = self.get_property(
            "no_of_samples_in_basket", SampleChangerMockup.NO_OF_SAMPLES_IN_BASKET
        )

        for i in range(self.no_of_baskets):
            basket = Container.Basket(
                self, i + 1, samples_num=self.no_of_samples_in_basket
            )
            self._add_component(basket)

        self._init_sc_contents()
        self.signal_wait_task = None
        AbstractSampleChanger.SampleChanger.init(self)

        self.log_filename = self.get_property("log_filename")

    def get_log_filename(self):
        return self.log_filename

    def load(self, sample, wait=False):
        self.emit("fsmConditionChanged", "sample_mounting_sample_changer", True)
        previous_sample = self.get_loaded_sample()
        self._set_state(AbstractSampleChanger.SampleChangerState.Loading)
        self._reset_loaded_sample()

        if isinstance(sample, tuple):
            basket, sample = sample
        else:
            basket, sample = sample.split(":")

        self._selected_basket = basket = int(basket)
        self._selected_sample = sample = int(sample)

        msg = "Loading sample %d:%d" % (basket, sample)
        logging.getLogger("user_level_log").info(
            "Sample changer: %s. Please wait..." % msg
        )

        self.emit("progressInit", (msg, 100))
        for step in range(2 * 100):
            self.emit("progressStep", int(step / 2.0))
            time.sleep(0.01)

        mounted_sample = self.get_component_by_address(
            Container.Pin.get_sample_address(basket, sample)
        )
        self._set_state(AbstractSampleChanger.SampleChangerState.Ready)

        if mounted_sample is not previous_sample:
            self._trigger_loaded_sample_changed_event(mounted_sample)
        self.update_info()
        logging.getLogger("user_level_log").info("Sample changer: Sample loaded")
        self.emit("progressStop", ())

        self.emit("fsmConditionChanged", "sample_is_loaded", True)
        self.emit("fsmConditionChanged", "sample_mounting_sample_changer", False)

        return self.get_loaded_sample()

    def unload(self, sample_slot=None, wait=None):
        logging.getLogger("user_level_log").info("Unloading sample")
        sample = self.get_loaded_sample()
        sample._set_loaded(False, True)
        self._selected_basket = -1
        self._selected_sample = -1
        self._trigger_loaded_sample_changed_event(self.get_loaded_sample())
        self.emit("fsmConditionChanged", "sample_is_loaded", False)

    def get_loaded_sample(self):
        return self.get_component_by_address(
            Container.Pin.get_sample_address(
                self._selected_basket, self._selected_sample
            )
        )

    def is_mounted_sample(self, sample):
        return (
            self.get_component_by_address(
                Container.Pin.get_sample_address(sample[0], sample[1])
            )
            == self.get_loaded_sample()
        )

    def _do_abort(self):
        return

    def _do_change_mode(self):
        return

    def _do_update_info(self):
        return

    def _do_select(self, component):
        return

    def _do_scan(self, component, recursive):
        return

    def _do_load(self, sample=None):
        return

    def _do_unload(self, sample_slot=None):
        return

    def _do_reset(self):
        return

    def _init_sc_contents(self):
        """
        Initializes the sample changer content with default values.

        :returns: None
        :rtype: None
        """
        named_samples = {}
        if self.has_object("test_sample_names"):
            for tag, val in self["test_sample_names"].get_properties().items():
                named_samples[val] = tag

        for basket_index in range(self.no_of_baskets):
            basket = self.get_components()[basket_index]
            datamatrix = None
            present = True
            scanned = False
            basket._set_info(present, datamatrix, scanned)

        sample_list = []
        for basket_index in range(self.no_of_baskets):
            for sample_index in range(self.no_of_samples_in_basket):
                sample_list.append(
                    (
                        "",
                        basket_index + 1,
                        sample_index + 1,
                        1,
                        Container.Pin.STD_HOLDERLENGTH,
                    )
                )
        for spl in sample_list:
            address = Container.Pin.get_sample_address(spl[1], spl[2])
            sample = self.get_component_by_address(address)
            sample_name = named_samples.get(address)
            if sample_name is not None:
                sample._name = sample_name
            datamatrix = "matr%d_%d" % (spl[1], spl[2])
            present = scanned = loaded = has_been_loaded = False
            sample._set_info(present, datamatrix, scanned)
            sample._set_loaded(loaded, has_been_loaded)
            sample._set_holder_length(spl[4])

        self._set_state(AbstractSampleChanger.SampleChangerState.Ready)

    def is_powered(self):
        return True



from Bliss import Bliss


class ID30Controller(Bliss):
    def __init__(self, *args):
        Bliss.__init__(self, *args)

    def init(self, *args):
        Bliss.init(self)

    def set_diagfile(self, diagfile):
        self.minidiff.diagfile = diagfile

    def __getattr__(self, attr):
        if attr.startswith("__") or attr == "minidiff":
            raise AttributeError(attr)
        return getattr(self.minidiff, attr)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Example xml file:
<device class="MotorMockupBis">
  <username>Detector Distance</username>
  <actuator_name>dtox</actuator_name>
  <tolerance>1e-2</tolerance>
</device>
"""

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MotorMockupBis(AbstractMotor):
    """Motor Motor implementation"""

    def init(self):
        """Initialise the motor"""
        super().init()
        # init state to match motor's one
        self.update_state(self.STATES.READY)

    def get_state_nu(self):
        """Get the motor state.
        Returns:
            (enum HardwareObjectState): Motor state.
        """
        return HardwareObjectState.READY

    def get_value(self):
        return self._nominal_value

    def _set_value(self, value):
        """Move motor to absolute value.
        Args:
            value (float): target value
        """
        self.update_state(self.STATES.BUSY)
        # self.update_value(value)
        self._nominal_value = value
        self.update_state(self.STATES.READY)



from mxcubecore.HardwareObjects import MiniDiff


class ID30A3MD2M(MiniDiff.MiniDiff):
    def __init__(self, name):
        MiniDiff.MiniDiff.__init__(self, name)
        MiniDiff.qmo.CentredPosition.set_diffractometer_motor_names(
            "phi", "focus", "phiz", "phiy", "zoom", "sampx", "sampy"
        )

    def init(self):
        self.controller = self.get_object_by_role("controller")
        MiniDiff.MiniDiff.init(self)

    def oscil(self, *args, **kwargs):
        self.controller.oscil(*args, **kwargs)

    def helical_oscil(self, *args, **kwargs):
        self.controller.helical_oscil(*args, **kwargs)



#! /usr/bin/env python
# encoding: utf-8
"""Workflow connection, interfacing to external workflow engine
using py4j and Abstract Beamline Interface messages

License:

This file is part of MXCuBE.

MXCuBE is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MXCuBE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with MXCuBE. If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import logging
import os
import signal
import socket  # Do NOT remove - see 'origsocket' below
import subprocess
import sys
import time
import uuid
from urllib.parse import urlparse

from py4j import (
    clientserver,
    java_gateway,
)
from py4j.protocol import Py4JJavaError

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectYaml
from mxcubecore.HardwareObjects.Gphl import GphlMessages
from mxcubecore.utils import conversion

# NB this is patching the original socket module in to avoid the
# monkeypatched version we get from gevent - that causes errors.
# It depends on knowing where in py4j socket is imported
# Hacky, but the best solution to making py4j and gevent compatible


origsocket = sys.modules.pop("socket")
_origsocket = sys.modules.pop("_socket")
import socket

java_gateway.socket = socket
clientserver.socket = socket
sys.modules["socket"] = origsocket
sys.modules["_socket"] = _origsocket
del origsocket
del _origsocket

try:
    # This file already does the alternative imports plus some tweaking
    # TODO It ought to be moved out as an accessible Util file, but meanwhile
    # Here we take care of the case where it is missing.
    from mxcubecore.dispatcher import dispatcher
except ImportError:
    try:
        from louie import dispatcher
    except ImportError:
        from pydispatch import (
            dispatcher,
            robustapply,
            saferef,
        )

        saferef.safe_ref = saferef.safeRef
        robustapply.robust_apply = robustapply.robustApply

__copyright__ = """ Copyright © 2016 - 2019 by Global Phasing Ltd. """
__license__ = "LGPLv3+"
__author__ = "Rasmus H Fogh"


class GphlWorkflowConnection(HardwareObjectYaml):
    """
    This HO acts as a gateway to the Global Phasing workflow engine.
    """

    def __init__(self, name):
        super().__init__(name)
        # Py4J gateway to external workflow program
        self._gateway = None
        self.msg_class_imported = False

        # ID for current workflow calculation
        self._enactment_id = None

        # Queue for communicating with MXCuBE HardwareObject
        self.workflow_queue = None
        self._await_result = None
        self._running_process = None
        self.collect_emulator_process = None

        # Configured parameters
        self.directory_locations = {}
        self.ssh_options = {}
        self.gphl_subdir = "GPHL"
        self.gphl_persistname = "persistence"
        self.connection_parameters = {}
        self.software_paths = {}
        self.software_properties = {}

        self.update_state(self.STATES.UNKNOWN)

    def init(self):
        super().init()

        # Adapt connections if we are running via ssh
        if self.ssh_options:
            self.connection_parameters["python_address"] = socket.gethostname()

        # Adapt paths and properties to use directory_locations
        locations = self.directory_locations
        installdir = locations["GPHL_INSTALLATION"]
        paths = self.software_paths
        properties = self.software_properties

        for tag, val in paths.items():
            val2 = val.format(**locations)
            if not os.path.isabs(val2):
                val2 = HWR.get_hardware_repository().find_in_repository(val)
                if val2 is None:
                    raise ValueError("File path %s not recognised" % val)
            paths[tag] = val2
        paths["GPHL_INSTALLATION"] = locations["GPHL_INSTALLATION"]
        paths["runworkflow"] = "%s/ASTRAWorkflows/bin/runworkflow" % installdir

        for tag, val in properties.items():
            val2 = val.format(**locations)
            if not os.path.isabs(val2):
                val2 = HWR.get_hardware_repository().find_in_repository(val)
                if val2 is None:
                    raise ValueError("File path %s not recognised" % val)
            paths[tag] = properties[tag] = val2

        # Set master location, based on known release directory structure
        properties["co.gphl.wf.bin"] = os.path.join(
            locations["GPHL_INSTALLATION"], "exe"
        )
        if "GPHL_XDS_PATH" in paths:
            properties["co.gphl.wf.xds.bin"] = os.path.join(
                paths["GPHL_XDS_PATH"], "xds_par"
            )

        self.update_state(self.STATES.OFF)

    def to_java_time(self, time_in):
        """Convert time in seconds since the epoch (python time) to Java time value"""
        return self._gateway.jvm.java.lang.Long(int(time_in * 1000))

    def get_executable(self, name):
        """Get location of executable binary for program called 'name'"""
        tag = "co.gphl.wf.%s.bin" % name
        result = self.software_paths.get(tag)
        if not result:
            result = os.path.join(self.software_paths["GPHL_INSTALLATION"], "exe", name)
        #
        return result

    def get_bdg_licence_dir(self, name):
        """Get directory containing specific licence file (if any)
        for program called 'name'"""
        tag = "co.gphl.wf.%s.bdg_licence_dir" % name
        result = self.software_paths.get(tag)
        #
        return result

    def open_connection(self):
        if self._gateway is None:
            logging.getLogger("HWR").debug("Opening GΦL connection")
        else:
            return

        params = self.connection_parameters

        python_parameters = {}
        val = params.get("python_address")
        if val is not None:
            python_parameters["address"] = val
        val = params.get("python_port")
        if val is not None:
            python_parameters["port"] = val

        java_parameters = {"auto_convert": True}
        val = params.get("java_address")
        if val is not None:
            java_parameters["address"] = val
        val = params.get("java_port")
        if val is not None:
            java_parameters["port"] = val

        logging.getLogger("HWR").debug(
            "Opening GΦL connection: %s ",
            (", ".join("%s:%s" % tt0 for tt0 in sorted(params.items()))),
        )

        self._gateway = clientserver.ClientServer(
            java_parameters=clientserver.JavaParameters(**java_parameters),
            python_parameters=clientserver.PythonParameters(**python_parameters),
            python_server_entry_point=self,
        )

    def start_workflow(self, workflow_queue, workflow_model_obj):
        # NBNB All command line option values are put in quotes (repr) when
        # the workflow is invoked remotely through ssh.

        if self.get_state() == self.STATES.UNKNOWN:
            logging.getLogger("HWR").warning(
                "GphlWorkflowConnection not correctly initialised - check for errors"
            )

        elif self.get_state() != self.STATES.OFF:
            # NB, for now workflow is started as the connection is made,
            # so we are never in state 'ON'/STANDBY
            raise RuntimeError("Workflow is already running, cannot be started")

        self.msg_class_imported = False

        # Cannot be done in init, where the api.sessions link is not yet ready
        self.software_paths["GPHL_WDIR"] = os.path.join(
            HWR.beamline.session.get_base_process_directory(), self.gphl_subdir
        )

        strategy_settings = workflow_model_obj.strategy_settings
        wf_settings = HWR.beamline.gphl_workflow.settings

        ssh_options = self.ssh_options
        in_shell = bool(ssh_options)
        if in_shell:
            ssh_options = ssh_options.copy()
            host = ssh_options.pop("Host")
            command_list = ["ssh"]
            if "ConfigFile" in ssh_options:
                command_list.extend(("-F", ssh_options.pop("ConfigFile")))
            for tag, val in sorted(ssh_options.items()):
                command_list.extend(("-o", "%s=%s" % (tag, val)))
            command_list.append(host)
        else:
            command_list = []
        runworkflow_opts = []
        command_list.append(self.software_paths["runworkflow"])

        # # HACK - debug options REMOVE!
        # import socket
        # sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # sock.connect(("8.8.8.8", 80))
        # ss0 = "-agentlib:jdwp=transport=dt_socket,address=%s:8050,server=y,suspend=y"
        # command_list.append(ss0 % sock.getsockname()[0])

        for tag, val in sorted(wf_settings.get("invocation_properties", {}).items()):
            runworkflow_opts.extend(
                conversion.java_property(tag, val, quote_value=in_shell)
            )

        init_spot_dir = workflow_model_obj.init_spot_dir
        if init_spot_dir:
            runworkflow_opts.extend(
                conversion.java_property("co.gphl.wf.initSpotDir", init_spot_dir)
            )

        # We must get hold of the options here, as we need wdir for a property
        workflow_options = dict(strategy_settings.get("options", {}))
        calibration_name = workflow_options.get("calibration")
        if calibration_name:
            # Expand calibration base name - to simplify identification.
            workflow_options["calibration"] = "%s_%s" % (
                calibration_name,
                workflow_model_obj.get_name(),
            )
        if workflow_model_obj.wftype in ("acquisition", "diffractcal"):
            workflow_options["strategy"] = workflow_model_obj.initial_strategy

        path_template = workflow_model_obj.get_path_template()
        if "prefix" in workflow_options:
            workflow_options["prefix"] = path_template.base_prefix
        workflow_options["wdir"] = self.software_paths["GPHL_WDIR"]
        workflow_options["persistname"] = self.gphl_persistname

        # Set the workflow root subdirectory parameter from the base image directory
        image_root = os.path.abspath(HWR.beamline.session.get_base_image_directory())
        if strategy_settings["wftype"] != "transcal":
            workflow_options["appdir"] = (
                HWR.beamline.session.get_base_process_directory()
            )
            rootsubdir = path_template.directory[len(image_root) :]
            if rootsubdir.startswith(os.path.sep):
                rootsubdir = rootsubdir[1:]
            if rootsubdir:
                workflow_options["rootsubdir"] = rootsubdir

        # Hardcoded - location for log output
        runworkflow_opts.extend(
            conversion.java_property(
                "co.gphl.wf.wdir", workflow_options["wdir"], quote_value=in_shell
            )
        )

        command_list.append(strategy_settings["wf_selection"])

        for keyword, value in wf_settings.get("workflow_properties", {}).items():
            command_list.extend(
                conversion.java_property(keyword, value, quote_value=in_shell)
            )
        for keyword, value in self.software_properties.items():
            command_list.extend(
                conversion.java_property(keyword, value, quote_value=in_shell)
            )

        for keyword, value in workflow_options.items():
            command_list.extend(
                conversion.command_option(keyword, value, quote_value=in_shell)
            )
        #
        wdir = workflow_options.get("wdir")
        # NB this creates the appdir as well (wdir is within appdir)
        if not os.path.isdir(wdir):
            try:
                os.makedirs(wdir)
            except:
                # No need to raise error - program will fail downstream
                logging.getLogger("HWR").error(
                    "Could not create GΦL working directory: %s", wdir
                )

        for ss0 in command_list:
            ss0 = ss0.rsplit("=", maxsplit=1)[-1]
            if ss0.startswith("/") and "*" not in ss0 and not os.path.exists(ss0):
                logging.getLogger("HWR").warning("File does not exist : %s", ss0)

        logging.getLogger("HWR").info("GΦL execute :\n%s", " ".join(command_list))

        # Get environmental variables
        envs = os.environ.copy()

        # # Trick to allow unauthorised account (e.g. ESRF: opid30) to run GPhL programs
        # # Any value is OK, just setting it is enough.
        # envs["AutoPROCWorkFlowUser"] = "1"

        # These env variables are needed in some cases for wrapper scripts
        # Specifically for the stratcal wrapper.
        envs["GPHL_INSTALLATION"] = self.software_paths["GPHL_INSTALLATION"]
        GPHL_XDS_PATH = self.software_paths.get("GPHL_XDS_PATH")
        if GPHL_XDS_PATH:
            envs["GPHL_XDS_PATH"] = GPHL_XDS_PATH
        GPHL_CCP4_PATH = self.software_paths.get("GPHL_CCP4_PATH")
        if GPHL_CCP4_PATH:
            envs["GPHL_CCP4_PATH"] = GPHL_CCP4_PATH
        GPHL_AUTOPROC_PATH = self.software_paths.get("GPHL_AUTOPROC_PATH")
        if GPHL_AUTOPROC_PATH:
            envs["GPHL_AUTOPROC_PATH"] = GPHL_AUTOPROC_PATH
        GPHL_MINICONDA_PATH = self.software_paths.get("GPHL_MINICONDA_PATH")
        if GPHL_MINICONDA_PATH:
            envs["GPHL_MINICONDA_PATH"] = GPHL_MINICONDA_PATH
        if runworkflow_opts:
            envs["RUNWORKFLOW_OPTS"] = " ".join(runworkflow_opts)

        logging.getLogger("HWR").debug(
            "Executing GΦL workflow, in environment %s", envs
        )
        try:
            self._running_process = subprocess.Popen(command_list, env=envs)
        except Exception:
            logging.getLogger().exception("Error in spawning workflow application")
            raise

        self.workflow_queue = workflow_queue
        logging.getLogger("py4j.clientserver").setLevel(logging.WARNING)
        self.update_state(self.STATES.READY)

        logging.getLogger("HWR").debug(
            "GΦL workflow pid, returncode : %s, %s"
            % (self._running_process.pid, self._running_process.returncode)
        )

    def workflow_ended(self):
        if self.get_state() == self.STATES.OFF:
            # No workflow to abort
            return

        logging.getLogger("HWR").debug("GΦL workflow ended")
        self.update_state(self.STATES.OFF)
        if self._await_result is not None:
            # We are awaiting an answer - give an abort
            self._await_result.append((GphlMessages.BeamlineAbort(), None))
            time.sleep(0.2)
        elif self._running_process is not None:
            self._running_process = None
            # NBNB TODO how do we close down the workflow if there is no answer pending?

        self._enactment_id = None
        self.workflow_queue = None
        self._await_result = None

        xx0 = self.collect_emulator_process
        if xx0 is not None:
            self.collect_emulator_process = "ABORTED"
            try:
                if xx0.poll() is None:
                    xx0.send_signal(signal.SIGINT)
                    time.sleep(3)
                    if xx0.poll() is None:
                        xx0.terminate()
                        time.sleep(9)
                        if xx0.poll() is None:
                            xx0.kill()
            except:
                logging.getLogger("HWR").info(
                    "Exception while terminating external workflow process %s", xx0
                )
                logging.getLogger("HWR").info("Error was:", exc_info=True)

    def close_connection(self):
        logging.getLogger("HWR").debug("GΦL Close connection ")
        xx0 = self._gateway
        self._gateway = None
        if xx0 is not None:
            try:
                # Exceptions 'can easily happen' (py4j docs)
                # Without raise_exception exceptions in the first part of the shutddown
                # will be caught and the rest of the shutdown will continue.
                # which is what we want.
                # xx0.shutdown(raise_exception=True)
                xx0.shutdown()
            except Exception:
                logging.getLogger("HWR").debug(
                    "Exception during py4j gateway shutdown. Ignored"
                )

    def abort_workflow(self, message=None):
        """Abort workflow - may be called from controller in any state"""

        logging.getLogger("HWR").info("Aborting workflow: %s", message)
        logging.getLogger("user_level_log").info("Aborting workflow ...")
        if self._await_result is not None:
            # Workflow waiting for answer - send abort
            self._await_result = [(GphlMessages.BeamlineAbort(), None)]

        # Shut down hardware object
        que = self.workflow_queue
        if que is None:
            self.workflow_ended()
        else:
            # If the queue is running,
            # workflow_ended will be called from post_execute
            que.put_nowait(StopIteration)

    def processText(self, py4j_message):
        """Receive and process info message from workflow server
        Return goes to server

        NB Callled freom external java) workflow"""
        xx0 = self._decode_py4j_message(py4j_message)
        message_type = xx0.message_type
        payload = xx0.payload
        correlation_id = xx0.correlation_id

        if not payload:
            logging.getLogger("HWR").warning(
                "GΦL Empty or unparsable information message. Ignored"
            )
        elif self.workflow_queue is not None:
            # Could happen if we have ended the workflow
            self.workflow_queue.put_nowait(
                (message_type, payload, correlation_id, None)
            )

    def processMessage(self, py4j_message):
        """Receive and process message from workflow server
        Return goes to server

        NB Callled freom external java) workflow"""
        if self.get_state() is self.STATES.OFF:
            return None

        if not self.msg_class_imported:
            # Guard agaiunst Java classes being moved to different packages
            try:
                msg_class = (
                    self._gateway.jvm.py4j.reflection.ReflectionUtil.classForName(
                        "co.gphl.sdcp.astra.service.py4j.Py4jMessage"
                    )
                )
                java_gateway.java_import(
                    self._gateway.jvm, "co.gphl.sdcp.astra.service.py4j.Py4jMessage"
                )
            except Py4JJavaError:
                msg_class = (
                    self._gateway.jvm.py4j.reflection.ReflectionUtil.classForName(
                        "co.gphl.sdcp.py4j.Py4jMessage"
                    )
                )
                java_gateway.java_import(
                    self._gateway.jvm, "co.gphl.sdcp.py4j.Py4jMessage"
                )
            logging.getLogger("HWR").debug(
                "GΦL workflow Py4jMessage class is: %s" % msg_class
            )
        self.msg_class_imported = True

        xx0 = self._decode_py4j_message(py4j_message)
        message_type = xx0.message_type
        payload = xx0.payload
        correlation_id = xx0.correlation_id

        if self._enactment_id is None:
            # NB this should be made less primitive
            # once we are past direct function calls
            self._enactment_id = xx0.enactment_id

        elif not payload:
            logging.getLogger("HWR").error(
                "GΦL message lacks payload - sending 'Abort' to external workflow"
            )
            return self._response_to_server(
                GphlMessages.BeamlineAbort(), correlation_id
            )

        if message_type in ("SubprocessStarted", "SubprocessStopped"):
            if self.workflow_queue is not None:
                # Could happen if we have ended the workflow
                self.workflow_queue.put_nowait(
                    (message_type, payload, correlation_id, None)
                )
            logging.getLogger("HWR").debug("Subprocess start/stop - return None")
            return None

        elif message_type in (
            "RequestConfiguration",
            "GeometricStrategy",
            "CollectionProposal",
            "ChooseLattice",
            "RequestCentring",
            "ObtainPriorInformation",
            "PrepareForCentring",
        ):
            # Requests:
            self._await_result = []
            self.update_state(self.STATES.BUSY)
            if self.workflow_queue is None:
                # Could be None if we have ended the workflow
                return self._response_to_server(
                    GphlMessages.BeamlineAbort(), correlation_id
                )
            else:
                self.workflow_queue.put_nowait(
                    (message_type, payload, correlation_id, self._await_result)
                )
                while not self._await_result:
                    time.sleep(0.1)
                result, correlation_id = self._await_result.pop(0)
                if self.get_state() == self.STATES.BUSY:
                    self.update_state(self.STATES.READY)
                self._await_result = None

                if result is StopIteration:
                    result = GphlMessages.BeamlineAbort()
                    self.workflow_queue.put_nowait(
                        (
                            "WorkflowAborted",
                            GphlMessages.WorkflowAborted(),
                            correlation_id,
                            None,
                        )
                    )
                    self.workflow_ended()
                else:
                    logging.getLogger("HWR").debug(
                        "GΦL - response=%s messageId=%s"
                        % (result.__class__.__name__, correlation_id)
                    )
                return self._response_to_server(result, correlation_id)

        elif message_type in ("WorkflowAborted", "WorkflowCompleted", "WorkflowFailed"):
            if self.workflow_queue is not None:
                # Could happen if we have ended the workflow
                self.workflow_queue.put_nowait(
                    (message_type, payload, correlation_id, None)
                )
                self.workflow_ended()
            logging.getLogger("HWR").debug("Aborting - return None")
            return None

        else:
            logging.getLogger("HWR").error(
                "GΦL Unknown message type: %s - aborting", message_type
            )
            return self._response_to_server(
                GphlMessages.BeamlineAbort(), correlation_id
            )

    # Conversion to Python

    def _decode_py4j_message(self, py4j_message):
        """Extract messageType and convert py4J object to python object"""

        # Determine message type
        message_type = py4j_message.getPayloadClass().getSimpleName()
        if message_type.endswith("Impl"):
            message_type = message_type[:-4]
        if message_type == "RequestConfiguration":
            xx0 = py4j_message.getPayload().getEnactmentId()
            enactment_id = xx0 and xx0.toString()
        else:
            enactment_id = None

        xx0 = py4j_message.getCorrelationId()
        correlation_id = xx0 and xx0.toString()
        if message_type == "String":
            payload = py4j_message.getPayload()

        else:
            logging.getLogger("HWR").debug(
                "GPhL incoming: message=%s, jobId=%s,  messageId=%s"
                % (message_type, enactment_id, correlation_id)
            )
            converterName = "_%s_to_python" % message_type

            try:
                # determine converter function
                converter = getattr(self, converterName)
            except AttributeError:
                logging.getLogger("HWR").error(
                    "GΦL Message type %s not recognised (no %s function)"
                    % (message_type, converterName)
                )
                payload = None
            else:
                try:
                    # Convert to Python objects
                    payload = converter(py4j_message.getPayload())
                except NotImplementedError:
                    logging.getLogger("HWR").error(
                        "Processing of GΦL message %s not implemented", message_type
                    )
                    payload = None
        #
        return GphlMessages.ParsedMessage(
            message_type, payload, enactment_id, correlation_id
        )

    def _RequestConfiguration_to_python(self, py4jRequestConfiguration):
        return GphlMessages.RequestConfiguration()

    def _ObtainPriorInformation_to_python(self, py4jObtainPriorInformation):
        return GphlMessages.ObtainPriorInformation()

    def _PrepareForCentring_to_python(self, py4jPrepareForCentring):
        return GphlMessages.PrepareForCentring()

    def _GeometricStrategy_to_python(self, py4jGeometricStrategy):
        uuidString = py4jGeometricStrategy.getId().toString()
        sweeps = frozenset(
            self._Sweep_to_python(x) for x in py4jGeometricStrategy.getSweeps()
        )
        beamSetting = py4jGeometricStrategy.getDefaultBeamSetting()
        if beamSetting:
            beamSetting = self._BeamSetting_to_python(beamSetting)
        else:
            beamSetting = None
        detectorSetting = py4jGeometricStrategy.getDefaultDetectorSetting()
        if detectorSetting:
            detectorSetting = self._DetectorSetting_to_python(detectorSetting)
        else:
            detectorSetting = None
        return GphlMessages.GeometricStrategy(
            # isInterleaved=py4jGeometricStrategy.isInterleaved(),
            isUserModifiable=py4jGeometricStrategy.isUserModifiable(),
            allowedWidths=py4jGeometricStrategy.getAllowedWidths(),
            sweepOffset=py4jGeometricStrategy.getSweepOffset(),
            sweepRepeat=py4jGeometricStrategy.getSweepRepeat(),
            defaultWidthIdx=py4jGeometricStrategy.getDefaultWidthIdx(),
            defaultBeamSetting=beamSetting,
            defaultDetectorSetting=detectorSetting,
            sweeps=sweeps,
            id_=uuid.UUID(uuidString),
        )

    def _SubprocessStarted_to_python(self, py4jSubprocessStarted):
        return GphlMessages.SubprocessStarted(name=py4jSubprocessStarted.getName())

    def _SubprocessStopped_to_python(self, py4jSubprocessStopped):
        return GphlMessages.SubprocessStopped()

    def _ChooseLattice_to_python(self, py4jChooseLattice):
        # NB the functions return different types, so toString is needed in only once
        indexingFormat = py4jChooseLattice.getIndexingFormat().toString()
        indexingHeader = py4jChooseLattice.getIndexingHeader()
        inputCell = py4jChooseLattice.getUserProvidedCell()
        userProvidedCell = self._UnitCell_to_python(inputCell) if inputCell else None
        return GphlMessages.ChooseLattice(
            indexingSolutions=tuple(
                self._IndexingSolution_to_python(sol)
                for sol in py4jChooseLattice.getIndexingSolutions()
            ),
            indexingFormat=indexingFormat,
            indexingHeader=indexingHeader,
            priorCrystalClasses=tuple(
                ccl.toString() for ccl in py4jChooseLattice.getPriorCrystalClasses()
            ),
            priorSpaceGroup=py4jChooseLattice.getPriorSpaceGroup(),
            priorSpaceGroupString=py4jChooseLattice.getPriorSpaceGroupString(),
            userProvidedCell=userProvidedCell,
        )

    def _CollectionProposal_to_python(self, py4jCollectionProposal):
        uuidString = py4jCollectionProposal.getId().toString()
        strategy = self._GeometricStrategy_to_python(
            py4jCollectionProposal.getStrategy()
        )
        text_type = conversion.text_type
        id2Sweep = dict((text_type(x.id_), x) for x in strategy.sweeps)
        scans = []
        for py4jScan in py4jCollectionProposal.getScans():
            sweep = id2Sweep[py4jScan.getSweep().getId().toString()]
            scans.append(self._Scan_to_python(py4jScan, sweep))
        return GphlMessages.CollectionProposal(
            relativeImageDir=py4jCollectionProposal.getRelativeImageDir(),
            strategy=strategy,
            scans=scans,
            id_=uuid.UUID(uuidString),
        )

    def __WorkflowDone_to_python(self, py4jWorkflowDone, cls):
        Issue = GphlMessages.Issue
        issues = []
        for py4jIssue in py4jWorkflowDone.getIssues():
            component = py4jIssue.getComponent()
            message = py4jIssue.getMessage()
            code = py4jIssue.getCode()
            issues.append(Issue(component=component, message=message, code=code))
        #
        return cls(issues=issues)

    def _WorkflowCompleted_to_python(self, py4jWorkflowCompleted):
        return self.__WorkflowDone_to_python(
            py4jWorkflowCompleted, GphlMessages.WorkflowCompleted
        )

    def _WorkflowAborted_to_python(self, py4jWorkflowAborted):
        return self.__WorkflowDone_to_python(
            py4jWorkflowAborted, GphlMessages.WorkflowAborted
        )

    def _WorkflowFailed_to_python(self, py4jWorkflowFailed):
        return self.__WorkflowDone_to_python(
            py4jWorkflowFailed, GphlMessages.WorkflowFailed
        )

    def _RequestCentring_to_python(self, py4jRequestCentring):
        goniostatRotation = self._GoniostatRotation_to_python(
            py4jRequestCentring.getGoniostatRotation()
        )
        return GphlMessages.RequestCentring(
            currentSettingNo=py4jRequestCentring.getCurrentSettingNo(),
            totalRotations=py4jRequestCentring.getTotalRotations(),
            goniostatRotation=goniostatRotation,
        )

    def _GoniostatRotation_to_python(self, py4jGoniostatRotation, isSweepSetting=False):
        if py4jGoniostatRotation is None:
            return None

        uuidString = py4jGoniostatRotation.getId().toString()
        axisSettings = py4jGoniostatRotation.getAxisSettings()
        if isSweepSetting:
            scanAxis = py4jGoniostatRotation.getScanAxis()
            result = GphlMessages.GoniostatSweepSetting(
                id_=uuid.UUID(uuidString), scanAxis=scanAxis, **axisSettings
            )
        else:
            result = GphlMessages.GoniostatRotation(
                id_=uuid.UUID(uuidString), **axisSettings
            )
        py4jGoniostatTranslation = py4jGoniostatRotation.getTranslation()
        if py4jGoniostatTranslation:
            translationAxisSettings = py4jGoniostatTranslation.getAxisSettings()
            translationUuidString = py4jGoniostatTranslation.getId().toString()
            # Next line creates Translation and links it to Rotation
            GphlMessages.GoniostatTranslation(
                id_=uuid.UUID(translationUuidString),
                rotation=result,
                **translationAxisSettings,
            )
        return result

    def _BeamstopSetting_to_python(self, py4jBeamstopSetting):
        if py4jBeamstopSetting is None:
            return None
        uuidString = py4jBeamstopSetting.getId().toString()
        axisSettings = py4jBeamstopSetting.getAxisSettings()
        #
        return GphlMessages.BeamstopSetting(id_=uuid.UUID(uuidString), **axisSettings)

    def _DetectorSetting_to_python(self, py4jDetectorSetting):
        if py4jDetectorSetting is None:
            return None
        uuidString = py4jDetectorSetting.getId().toString()
        axisSettings = py4jDetectorSetting.getAxisSettings()
        #
        return GphlMessages.DetectorSetting(id_=uuid.UUID(uuidString), **axisSettings)

    def _BeamSetting_to_python(self, py4jBeamSetting):
        if py4jBeamSetting is None:
            return None
        uuidString = py4jBeamSetting.getId().toString()
        #
        return GphlMessages.BeamSetting(
            id_=uuid.UUID(uuidString), wavelength=py4jBeamSetting.getWavelength()
        )

    def _GoniostatSweepSetting_to_python(self, py4jGoniostatSweepSetting):
        return self._GoniostatRotation_to_python(
            py4jGoniostatSweepSetting, isSweepSetting=True
        )

    def _UnitCell_to_python(self, py4jUnitCell):
        cell_params = tuple(py4jUnitCell.getLengths()) + tuple(py4jUnitCell.getAngles())
        return GphlMessages.UnitCell(*cell_params)

    def _IndexingSolution_to_python(self, py4jIndexingSolution):
        return GphlMessages.IndexingSolution(
            bravaisLattice=py4jIndexingSolution.getBravaisLattice(),
            cell=self._UnitCell_to_python(py4jIndexingSolution.getCell()),
            isConsistent=py4jIndexingSolution.isConsistent(),
            latticeCharacter=py4jIndexingSolution.getLatticeCharacter(),
            qualityOfFit=py4jIndexingSolution.getQualityOfFit(),
        )

    def _Sweep_to_python(self, py4jSweep):
        # NB scans are not set - where scans are present in a message,
        # the link is set from the Scan side.

        uuidString = py4jSweep.getId().toString()
        return GphlMessages.Sweep(
            goniostatSweepSetting=self._GoniostatSweepSetting_to_python(
                py4jSweep.getGoniostatSweepSetting()
            ),
            detectorSetting=self._DetectorSetting_to_python(
                py4jSweep.getDetectorSetting()
            ),
            beamSetting=self._BeamSetting_to_python(py4jSweep.getBeamSetting()),
            start=py4jSweep.getStart(),
            width=py4jSweep.getWidth(),
            beamstopSetting=self._BeamstopSetting_to_python(
                py4jSweep.getBeamstopSetting()
            ),
            sweepGroup=py4jSweep.getSweepGroup(),
            id_=uuid.UUID(uuidString),
        )

    def _ScanExposure_to_python(self, py4jScanExposure):
        uuidString = py4jScanExposure.getId().toString()
        return GphlMessages.ScanExposure(
            time=py4jScanExposure.getTime(),
            transmission=py4jScanExposure.getTransmission(),
            id_=uuid.UUID(uuidString),
        )

    def _ScanWidth_to_python(self, py4jScanWidth):
        uuidString = py4jScanWidth.getId().toString()
        return GphlMessages.ScanWidth(
            imageWidth=py4jScanWidth.getImageWidth(),
            numImages=py4jScanWidth.getNumImages(),
            id_=uuid.UUID(uuidString),
        )

    def _Scan_to_python(self, py4jScan, sweep):
        uuidString = py4jScan.getId().toString()
        return GphlMessages.Scan(
            width=self._ScanWidth_to_python(py4jScan.getWidth()),
            exposure=self._ScanExposure_to_python(py4jScan.getExposure()),
            imageStartNum=py4jScan.getImageStartNum(),
            start=py4jScan.getStart(),
            sweep=sweep,
            filenameParams=py4jScan.getFilenameParams(),
            id_=uuid.UUID(uuidString),
        )

    # Conversion to Java

    def _payload_to_java(self, payload):
        """Convert Python payload object to java"""

        payloadType = payload.__class__.__name__

        if payloadType == "ConfigurationData":
            return self._ConfigurationData_to_java(payload)

        if payloadType == "BeamlineAbort":
            return self._BeamlineAbort_to_java(payload)

        if payloadType == "ReadyForCentring":
            return self._ReadyForCentring_to_java(payload)

        if payloadType == "SampleCentred":
            return self._SampleCentred_to_java(payload)

        if payloadType == "CollectionDone":
            # self.test_lattice_selection()
            return self._CollectionDone_to_java(payload)

        if payloadType == "SelectedLattice":
            return self._SelectedLattice_to_java(payload)

        if payloadType == "CentringDone":
            return self._CentringDone_to_java(payload)

        if payloadType == "PriorInformation":
            return self._PriorInformation_to_java(payload)

        raise ValueError(
            "Payload %s not supported for conversion to java" % payloadType
        )

    def _response_to_server(self, payload, correlation_id):
        """Create py4j message from py4j wrapper and current ids"""

        if correlation_id is not None:
            correlation_id = self._gateway.jvm.java.util.UUID.fromString(correlation_id)

        py4j_payload = self._payload_to_java(payload)

        try:
            response = self._gateway.jvm.co.gphl.sdcp.astra.service.py4j.Py4jMessage(
                py4j_payload, correlation_id
            )
        except:
            self.abort_workflow(
                message="Error creating Java message (%s) to send to workflow"
                % py4j_payload.getClass().getSimpleName()
            )
            return None
        else:
            return response

    def _CentringDone_to_java(self, centringDone):
        jvm = self._gateway.jvm
        return jvm.astra.messagebus.messages.information.CentringDoneImpl(
            jvm.co.gphl.beamline.v2_unstable.instrumentation.CentringStatus.valueOf(
                centringDone.status
            ),
            self.to_java_time(centringDone.timestamp),
            self._GoniostatTranslation_to_java(centringDone.goniostatTranslation),
        )

    def _ConfigurationData_to_java(self, configurationData):
        jvm = self._gateway.jvm
        return jvm.astra.messagebus.messages.information.ConfigurationDataImpl(
            self._gateway.jvm.java.io.File(configurationData.location)
        )

    def _ReadyForCentring_to_java(self, readyForCentring):
        return (
            self._gateway.jvm.astra.messagebus.messages.control.ReadyForCentringImpl()
        )

    def _PriorInformation_to_java(self, priorInformation):
        jvm = self._gateway.jvm
        buildr = jvm.astra.messagebus.messages.information.PriorInformationImpl.Builder(
            jvm.java.util.UUID.fromString(
                conversion.text_type(priorInformation.sampleId)
            )
        )
        xx0 = priorInformation.sampleName
        if xx0:
            buildr = buildr.sampleName(xx0)
        xx0 = priorInformation.rootDirectory
        if xx0:
            buildr = buildr.rootDirectory(xx0)
        buildr = buildr.userProvidedInfo(
            self._UserProvidedInfo_to_java(priorInformation.userProvidedInfo)
        )
        #
        return buildr.build()

    def _SampleCentred_to_java(self, sampleCentred):
        cls = self._gateway.jvm.astra.messagebus.messages.information.SampleCentredImpl

        # if sampleCentred.interleaveOrder:
        result = cls(
            float(sampleCentred.imageWidth),
            int(sampleCentred.wedgeWidth),
            float(sampleCentred.exposure),
            float(sampleCentred.transmission),
            list(sampleCentred.interleaveOrder),
            list(self._PhasingWavelength_to_java(x) for x in sampleCentred.wavelengths),
            self._BcsDetectorSetting_to_java(sampleCentred.detectorSetting),
            sampleCentred.repetition_count,
        )

        beamstopSetting = sampleCentred.beamstopSetting
        if beamstopSetting is not None:
            result.setBeamstopSetting(self._BeamstopSetting_to_java(beamstopSetting))

        translationSettings = sampleCentred.goniostatTranslations
        if translationSettings:
            result.setGoniostatTranslations(
                list(self._GoniostatTranslation_to_java(x) for x in translationSettings)
            )
        #
        return result

    def _CollectionDone_to_java(self, collectionDone):
        jvm = self._gateway.jvm
        proposalId = jvm.java.util.UUID.fromString(
            conversion.text_type(collectionDone.proposalId)
        )
        centrings = set(
            self._GoniostatTranslation_to_java(translation)
            for translation in collectionDone.centrings
        )
        scanIdMap = {}
        for item in collectionDone.scanIdMap.items():
            scanIdMap[jvm.java.util.UUID.fromString(conversion.text_type(item[0]))] = (
                jvm.java.util.UUID.fromString(conversion.text_type(item[1]))
            )
        return jvm.astra.messagebus.messages.information.CollectionDoneImpl(
            proposalId,
            collectionDone.status,
            collectionDone.procWithLatticeParams,
            scanIdMap,
            centrings,
        )

    def _SelectedLattice_to_java(self, selectedLattice):
        jvm = self._gateway.jvm

        builder = jvm.astra.messagebus.messages.information.SelectedLatticeImpl.Builder(
            self._IndexingSolution_to_java(selectedLattice.solution)
        )
        builder = builder.strategyDetectorSetting(
            self._BcsDetectorSetting_to_java(selectedLattice.strategyDetectorSetting)
        )
        builder = builder.strategyWavelength(
            self._PhasingWavelength_to_java(selectedLattice.strategyWavelength)
        )
        builder = builder.strategyControl(selectedLattice.strategyControl)
        builder = builder.userSpaceGroup(selectedLattice.userSpaceGroup)

        crystal_classes = selectedLattice.userCrystalClasses
        if crystal_classes:
            ccset = set(
                jvm.co.gphl.beamline.v2_unstable.domain_types.CrystalClass.fromStringList(
                    self.toJStringArray(crystal_classes)
                )
            )
            builder = builder.userCrystalClasses(ccset)
        urlstrings = selectedLattice.referenceReflectionFiles
        for urlstring in urlstrings:
            urltpl = urlparse(urlstring)
            host = urltpl.hostname
            port = urltpl.port or None
            if host:
                builder = builder.referenceFile(
                    urltpl.scheme, urltpl.hostname, port, urltpl.path
                )
            elif urltpl:
                builder = builder.referenceFile(urltpl.scheme, urltpl.path)
        #
        return builder.build()

    def _IndexingSolution_to_java(self, indexingSolution):
        jvm = self._gateway.jvm
        cell = indexingSolution.cell
        cell = cell and self._UnitCell_to_java(cell)
        result = jvm.astra.messagebus.messages.information.IndexingSolutionImpl(
            indexingSolution.bravaisLattice,
            indexingSolution.latticeCharacter,
            indexingSolution.isConsistent,
            indexingSolution.qualityOfFit,
            cell,
        )
        #
        return result

    def _BeamlineAbort_to_java(self, beamlineAbort):
        return (
            self._gateway.jvm.astra.messagebus.messages.instructions.BeamlineAbortImpl()
        )

    def _UserProvidedInfo_to_java(self, userProvidedInfo):
        jvm = self._gateway.jvm

        if userProvidedInfo is None:
            return None

        builder = (
            jvm.astra.messagebus.messages.information.UserProvidedInfoImpl.Builder()
        )

        for scatterer in userProvidedInfo.scatterers:
            builder = builder.addScatterer(self._AnomalousScatterer_to_java(scatterer))
        crystal_classes = userProvidedInfo.crystalClasses
        if crystal_classes:
            ccset = set(
                jvm.co.gphl.beamline.v2_unstable.domain_types.CrystalClass.fromStringList(
                    self.toJStringArray(crystal_classes)
                )
            )
            builder = builder.crystalClasses(ccset)
        xx0 = userProvidedInfo.spaceGroup
        if xx0:
            builder = builder.spaceGroup(xx0)
        xx0 = userProvidedInfo.spaceGroupString
        if xx0:
            builder = builder.spaceGroupString(xx0)
        xx0 = userProvidedInfo.cell
        if xx0 is not None:
            builder = builder.cell(self._UnitCell_to_java(xx0))
        if userProvidedInfo.expectedResolution:
            builder = builder.expectedResolution(
                float(userProvidedInfo.expectedResolution)
            )
        xx0 = userProvidedInfo.isAnisotropic
        if xx0 is not None:
            builder = builder.anisotropic(xx0)
        return builder.build()

    def _AnomalousScatterer_to_java(self, anomalousScatterer):
        jvm = self._gateway.jvm

        if anomalousScatterer is None:
            return None

        element = jvm.co.gphl.beamline.v2_unstable.domain_types.ChemicalElement.valueOf(
            anomalousScatterer.element
        )
        edge = jvm.co.gphl.beamline.v2_unstable.domain_types.AbsorptionEdge.valueOf(
            anomalousScatterer.edge
        )
        return jvm.astra.messagebus.messages.domain_types.AnomalousScattererImpl(
            element, edge
        )

    def _UnitCell_to_java(self, unitCell):
        if unitCell is None:
            return None

        lengths = [float(x) for x in unitCell.lengths]
        angles = [float(x) for x in unitCell.angles]
        return self._gateway.jvm.astra.messagebus.messages.domain_types.UnitCellImpl(
            lengths[0], lengths[1], lengths[2], angles[0], angles[1], angles[2]
        )

    def _PhasingWavelength_to_java(self, phasingWavelength):
        jvm = self._gateway.jvm

        if phasingWavelength is None:
            return None

        javaUuid = self._gateway.jvm.java.util.UUID.fromString(
            conversion.text_type(phasingWavelength.id_)
        )
        return jvm.astra.messagebus.messages.information.PhasingWavelengthImpl(
            javaUuid, float(phasingWavelength.wavelength), phasingWavelength.role
        )

    def _BcsDetectorSetting_to_java(self, bcsDetectorSetting):
        jvm = self._gateway.jvm

        if bcsDetectorSetting is None:
            return None

        orgxy = bcsDetectorSetting.orgxy
        # Need (temporarily?) because there is a primitive array, not a list,
        # on the other side
        orgxy_array = self._gateway.new_array(jvm.double, 2)
        orgxy_array[0] = orgxy[0]
        orgxy_array[1] = orgxy[1]
        axisSettings = dict(
            ((x, float(y)) for x, y in bcsDetectorSetting.axisSettings.items())
        )
        javaUuid = jvm.java.util.UUID.fromString(
            conversion.text_type(bcsDetectorSetting.id_)
        )
        return jvm.astra.messagebus.messages.instrumentation.BcsDetectorSettingImpl(
            float(bcsDetectorSetting.resolution), orgxy_array, axisSettings, javaUuid
        )

    def _GoniostatTranslation_to_java(self, goniostatTranslation):
        jvm = self._gateway.jvm

        if goniostatTranslation is None:
            return None

        gts = goniostatTranslation
        javaUuid = jvm.java.util.UUID.fromString(conversion.text_type(gts.id_))
        javaRotationId = jvm.java.util.UUID.fromString(
            conversion.text_type(gts.requestedRotationId)
        )
        axisSettings = dict(((x, float(y)) for x, y in gts.axisSettings.items()))
        newRotation = gts.newRotation
        if newRotation:
            if isinstance(newRotation, GphlMessages.GoniostatSweepSetting):
                javaNewRotation = self._GoniostatSweepSetting_to_java(newRotation)
            else:
                javaNewRotation = self._GoniostatRotation_to_java(newRotation)

            return (
                jvm.astra.messagebus.messages.instrumentation.GoniostatTranslationImpl(
                    axisSettings, javaUuid, javaRotationId, javaNewRotation
                )
            )
        else:
            return (
                jvm.astra.messagebus.messages.instrumentation.GoniostatTranslationImpl(
                    axisSettings, javaUuid, javaRotationId
                )
            )

    def _GoniostatRotation_to_java(self, goniostatRotation):
        jvm = self._gateway.jvm

        if goniostatRotation is None:
            return None

        grs = goniostatRotation
        javaUuid = jvm.java.util.UUID.fromString(conversion.text_type(grs.id_))
        axisSettings = dict(((x, float(y)) for x, y in grs.axisSettings.items()))
        # Long problematic, but now fixed (on both sides)
        return jvm.astra.messagebus.messages.instrumentation.GoniostatRotationImpl(
            axisSettings, javaUuid
        )

    def _GoniostatSweepSetting_to_java(self, goniostatSweepSetting):
        """Not currently in use, as you cannot replace SweepSettings,
        but may come back in if something changes"""
        jvm = self._gateway.jvm

        if goniostatSweepSetting is None:
            return None

        gss = goniostatSweepSetting
        javaUuid = jvm.java.util.UUID.fromString(conversion.text_type(gss.id_))
        axisSettings = dict(((x, float(y)) for x, y in gss.axisSettings.items()))
        return jvm.astra.messagebus.messages.instrumentation.GoniostatSweepSettingImpl(
            axisSettings, javaUuid, goniostatSweepSetting.scanAxis
        )

    def _BeamstopSetting_to_java(self, beamStopSetting):
        jvm = self._gateway.jvm

        if beamStopSetting is None:
            return None

        javaUuid = jvm.java.util.UUID.fromString(
            conversion.text_type(beamStopSetting.id_)
        )
        axisSettings = dict(
            ((x, float(y)) for x, y in beamStopSetting.axisSettings.items())
        )
        return jvm.astra.messagebus.messages.instrumentation.BeamstopSettingImpl(
            axisSettings, javaUuid
        )

    def toJStringArray(self, arr):
        """Modified from
        https://stackoverflow.com/questions/61230680/pyspark-py4j-create-java-string-array
        """
        jarr = self._gateway.new_array(self._gateway.jvm.java.lang.String, len(arr))
        for ind, val in enumerate(arr):
            jarr[ind] = val
        return jarr

    class Java:
        implements = ["co.gphl.py4j.PythonListener"]
